# packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from typing import Optional, Union
from enum import Enum

# qis
from qis.utils.df_melt import melt_scatter_data_with_xvar
import qis.perfstats.returns as ret
from qis.plots.scatter import plot_scatter
from qis.perfstats.config import ReturnTypes


def plot_returns_scatter(prices: pd.DataFrame,
                         benchmark: str = None,
                         benchmark_prices: Union[pd.Series, pd.DataFrame] = None,
                         freq: str = 'Q',
                         order: int = 2,
                         ci: Optional[int] = 95,
                         add_45line: bool = False,
                         is_vol_norm: bool = False,
                         y_column: str = 'Strategy returns',
                         xlabel: str = None,
                         ylabel: str = 'returns',
                         var_format: str = '{:.1%}',
                         title: Union[str, None] = None,
                         add_hue_model_label: bool = True,
                         hue_name: str = 'hue',
                         return_type: ReturnTypes = ReturnTypes.RELATIVE,
                         ax: plt.Subplot = None,
                         **kwargs
                         ) -> plt.Figure:

    if benchmark_prices is None:
        price_data_full = prices
    else:
        if isinstance(benchmark_prices, pd.Series):  # use benchmark set by series
            price_data_full = pd.concat([benchmark_prices, prices], axis=1)
            benchmark = benchmark_prices.name
            benchmark_prices = None
        else:  # for df price data must be sries
            if not isinstance(prices, pd.Series):
                raise ValueError(f"must be series\n{prices}")
            price_data_full = pd.concat([prices, benchmark_prices], axis=1)

    returns = ret.to_returns(prices=price_data_full,
                             include_start_date=True,
                             include_end_date=True,
                             return_type=return_type,
                             freq=freq)
    if is_vol_norm:
        returns = returns.divide(np.nanstd(returns, axis=0), axis=1)

    if benchmark_prices is None:
        scatter_data = melt_scatter_data_with_xvar(df=returns,
                                                   xvar_str=benchmark,
                                                   y_column=y_column,
                                                   hue_name=hue_name)
    else:
        scatter_data = melt_scatter_data_with_xvar(df=returns,
                                                   xvar_str=str(prices.name),
                                                   y_column=y_column,
                                                   hue_name=hue_name)
        benchmark = y_column
        y_column = str(prices.name)

    fig = plot_scatter(df=scatter_data,
                       x=benchmark,
                       y=y_column,
                       xlabel=xlabel or benchmark,
                       ylabel=ylabel,
                       hue=hue_name,
                       xvar_format=var_format,
                       yvar_format=var_format,
                       add_universe_model_label=False,
                       add_universe_model_prediction=False,
                       add_universe_model_ci=False,
                       add_hue_model_label=add_hue_model_label,
                       add_45line=add_45line,
                       title=title,
                       order=order,
                       ci=ci,
                       ax=ax,
                       **kwargs)
    return fig


class UnitTests(Enum):
    RETURNS = 1
    RETURNS2 = 2


def run_unit_test(unit_test: UnitTests):

    from qis.test_data import load_etf_data
    prices = load_etf_data().dropna()

    if unit_test == UnitTests.RETURNS:
        fig, ax = plt.subplots(1, 1, figsize=(8, 6))
        global_kwargs = dict(fontsize=8, linewidth=0.5, weight='normal', markersize=1)
        plot_returns_scatter(prices=prices,
                             benchmark='SPY',
                             var_format='{:.2%}',
                             ax=ax,
                             **global_kwargs)

    elif unit_test == UnitTests.RETURNS2:

        fig, ax = plt.subplots(1, 1, figsize=(8, 6))
        global_kwargs = dict(fontsize=8, linewidth=0.5, weight='normal', markersize=1)

        plot_returns_scatter(prices=prices[['SPY', 'TLT']],
                             benchmark='TLT',
                             y_column='benchmarks',
                             ylabel='SPY',
                             var_format='{:.2%}',
                             ax=ax,
                             **global_kwargs)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.RETURNS

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
analytics for dataframe aggregation
"""
import numpy as np
import pandas as pd
from scipy import stats
from typing import Union, Callable, List, Optional, Tuple
from enum import Enum

# qis
import qis.utils.np_ops as npo


def nanmean(df: pd.DataFrame) -> pd.Series:
    data_np = npo.to_finite_np(data=df, fill_value=np.nan)
    nanmean_data = pd.Series(data=np.nanmean(data_np, axis=1), index=df.index, name='nanmean')
    return nanmean_data


def nanmedian(df: pd.DataFrame) -> pd.Series:
    data_np = npo.to_finite_np(data=df, fill_value=np.nan)
    nanmean_data = pd.Series(data=np.nanmedian(data_np, axis=1), index=df.index, name='nanmedian')
    return nanmean_data


def nansum(df: pd.DataFrame) -> pd.Series:
    data_np = npo.to_finite_np(data=df, fill_value=np.nan)
    nansum_data = pd.Series(data=np.nansum(data_np, axis=1), index=df.index, name='nansum')
    return nansum_data


def nansum_positive(df: pd.DataFrame) -> pd.Series:
    signed_np_data = get_signed_np_data(df=df, is_positive=True)
    nansum_data = pd.Series(data=np.nansum(signed_np_data, axis=1), index=df.index, name='nansum_positive')
    return nansum_data


def nanmean_positive(df: pd.DataFrame) -> pd.Series:
    signed_np_data = get_signed_np_data(df=df, is_positive=True)
    nanmean_data = pd.Series(data=np.nanmean(signed_np_data, axis=1), index=df.index, name='nanmean_positive')
    return nanmean_data


def nansum_clip(df: pd.DataFrame,
                a_min: Optional[float] = None,
                a_max: Optional[float] = None,
                is_min_max_clip_fill: bool = True  # will be filled at min or max
                ) -> pd.Series:
    data_np = npo.to_finite_np(data=df, fill_value=np.nan, a_min=a_min, a_max=a_max, is_min_max_clip_fill=is_min_max_clip_fill)
    nanmean_data = pd.Series(data=np.nansum(data_np, axis=1), index=df.index, name='nansum_clip')
    return nanmean_data


def nanmean_clip(df: pd.DataFrame,
                 a_min: Optional[float] = None,
                 a_max: Optional[float] = None,
                 is_min_max_clip_fill: bool = True  # will be filled at min or max
                 ) -> pd.Series:
    data_np = npo.to_finite_np(data=df, fill_value=np.nan, a_min=a_min, a_max=a_max, is_min_max_clip_fill=is_min_max_clip_fill)
    nanmean_data = pd.Series(data=np.nanmean(data_np, axis=1), index=df.index, name='nanmean_clip')
    return nanmean_data


def nansum_negative(df: pd.DataFrame) -> pd.Series:
    signed_np_data = get_signed_np_data(df=df, is_positive=False)
    nansum_data = pd.Series(data=np.nansum(signed_np_data, axis=1), index=df.index, name='nansum_negative')
    return nansum_data


def abssum(df: pd.DataFrame) -> pd.Series:
    data_np = npo.to_finite_np(data=df, fill_value=np.nan)
    nansum_data = pd.Series(data=np.nansum(np.abs(data_np), axis=1), index=df.index, name='abssum')
    return nansum_data


def abssum_positive(df: pd.DataFrame) -> pd.Series:
    signed_np_data = get_signed_np_data(df=df, is_positive=True)
    nansum_data = pd.Series(data=np.nansum(np.abs(signed_np_data), axis=1), index=df.index, name='abssum_positive')
    return nansum_data


def abssum_negative(df: pd.DataFrame) -> pd.Series:
    signed_np_data = get_signed_np_data(df=df, is_positive=False)
    nansum_data = pd.Series(data=np.nansum(np.abs(signed_np_data), axis=1), index=df.index, name='abssum_negative')
    return nansum_data


def sum_weighted(df: pd.Series, weights: pd.Series) -> float:
    return np.nansum(np.multiply(df, weights))


def get_signed_np_data(df: pd.DataFrame,
                       is_positive: bool = True
                       ) -> np.ndarray:
    a = npo.to_finite_np(data=df, fill_value=np.nan)
    if is_positive:
        signed_np_data = np.where(np.greater(a, 0.0), a, np.nan)
    else:
        signed_np_data = np.where(np.less(a, 0.0), a, np.nan)
    return signed_np_data


def agg_median_mad(df: pd.DataFrame,
                   median_col: str = 'Median',
                   mad_col: str = 'Mad std',
                   ratio_col: str = 'Mad std %',
                   is_zeros_to_nan: bool = True,
                   scale: float = 0.67449
                   ) -> Tuple[pd.Series, pd.Series, pd.Series]:

    np_data = df.to_numpy()
    if is_zeros_to_nan:
        np_data = np.where(np.isclose(np_data, 0.0), np.nan, np_data)
    median = np.nanmedian(np_data, axis=1)
    mad = stats.median_abs_deviation(np_data, axis=1, nan_policy='omit', scale=scale)
    ratio = np.divide(mad, median, where=np.isfinite(median))
    median = pd.Series(median, index=df.index, name=median_col)
    mad = pd.Series(mad, index=df.index, name=mad_col)
    ratio = pd.Series(ratio, index=df.index, name=ratio_col)
    return median, mad, ratio


def agg_data_by_axis(df: pd.DataFrame,
                     total_column: Union[str, None] = None,
                     is_total_column_first: bool = False,  # last of not
                     agg_func: Callable[[pd.DataFrame], pd.Series] = np.nansum,
                     agg_total_func: Callable[[pd.Series], pd.Series] = np.nansum,
                     axis: int = 0
                     ) -> pd.Series:
    """
    take pandas data which are index=time series, column = assets
    agg_func: pd.DataFrame -> Union[pd.Series, pd.DataFrame] is function
    aggregation by axis=0 -> pd.Series[columns]
    aggregation by axis=1 -> pd.Series[index]
    """
    agg_data = pd.Series(data=agg_func(df, axis=axis), index=df.columns)
    # insert total
    if total_column is not None:
        if not isinstance(total_column, str):
            raise TypeError(f"in agg_data_by_groups: total_column must be string")
        # agg sum by columns
        agg_total = pd.Series(data=agg_total_func(agg_data), index=[total_column])

        if is_total_column_first == 0:
            agg_data = agg_total.append(agg_data)
        else:
            agg_data = agg_data.append(agg_total)

    return agg_data


def agg_dfs(dfs: List[pd.DataFrame],
            agg_func: Callable = np.nanmean
            ) -> pd.DataFrame:
    """
    compute average of same shaped pandas
    index must be the same
    not very efficient
    """
    # create pandas indexed by index*column with columns = len(datas)
    pd_data = pd.concat([df.stack() for df in dfs], axis=1)
    # apply mean to aggregate columns
    pd_avg = pd_data.apply(lambda x: agg_func(x.to_numpy()), axis=1)

    # transfrom to dataframe of original index and columns
    avg_data = pd_avg.unstack()

    return avg_data


class UnitTests(Enum):
    STACK = 1
    NAN_MEAN = 2
    TEST3 = 3


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.STACK:
        df = pd.DataFrame(data=[[0, 1], [2, 3]],
                          index=['cat', 'dog'],
                          columns=['weight', 'height'])
        print(df)
        stacked = df.stack()
        print(stacked)
        print(type(stacked.index))

        # melt to
        melted = pd.melt(df, value_vars=df.columns, var_name='myVarname', value_name='myValname')
        print(melted)

    elif unit_test == UnitTests.NAN_MEAN:

        # 4 * 3 matrix
        a = np.array([[np.nan, np.nan, np.nan],
                      [np.nan, np.nan, 1.0],
                      [np.nan, 2.0, 2.0],
                      [2.0, 3.0, 4.0]])

        #print(f"a={a[0]}\nmean(axis=0)={nan_func_to_data(a=a[0], func=np.nanmean, axis=0)};")

        print(f"a={a}\nmean(axis=0)={npo.nan_func_to_data(a=a, func=npo.np_nanmean, axis=0)};")

        print(f"a={a}\nmean(axis=1)={npo.nan_func_to_data(a=a, func=npo.np_nanmean, axis=1)};")

        pd_a = pd.DataFrame(a)
        print(f"pd_a={pd_a};")

        lambda_mean = pd_a.apply(lambda x: x.mean(), axis=1)
        # pd_a_pd = pd_a.apply(lambda x: np.nanmean(x), axis=1)
        print(f"lambda_mean\n{lambda_mean};")

        lambda_std = pd_a.apply(lambda x: x.std(), axis=1)
        # pd_a_pd = pd_a.apply(lambda x: np.nanstd(x), axis=1)
        print(f"lambda_std\n{lambda_std};")

    elif unit_test == UnitTests.TEST3:
        df = pd.DataFrame({'Date': ['2015-05-08', '2015-05-07', '2015-05-06', '2015-05-05', '2015-05-08', '2015-05-07',
                                    '2015-05-06', '2015-05-05'],
                           'Sym': ['aapl', 'aapl', 'aapl', 'aapl', 'aaww', 'aaww', 'aaww', 'aaww'],
                           'Data2': [11, 8, 10, 15, 110, 60, 100, 40],
                           'Data3': [5, 8, 6, 1, 50, 100, 60, 120]})
        print(df)
        df['Data4'] = df['Data3'].groupby(df['Date']).transform('sum')
        print(df)


if __name__ == '__main__':

    unit_test = UnitTests.TEST3

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)




"""
implement winsorizing of time series data using ewm
1. compute mean_t and vol_t
2. select x% of outliers defined by normalized score (x_t-mean_t) / vol_t
3. replace or trim outliers as specified
"""
# packages
import numpy as np
import pandas as pd
from enum import Enum

# qis
import qis.utils.dates as da
import qis.models.linear.ewm as ewm


class ConvolutionType(Enum):
    AUTO_CORR = 1
    SIGNAL_CORR = 2
    SIGNAL_BETA = 3


class SignalAggType(Enum):
    LAST_VALUE = 1
    MEAN = 2


def ewm_xy_convolution(returns: pd.DataFrame,
                       freq: str,
                       signals: pd.DataFrame = None,
                       convolution_type: ConvolutionType = ConvolutionType.AUTO_CORR,
                       signal_agg_type: SignalAggType = SignalAggType.LAST_VALUE,
                       is_ra_return: bool = False,
                       estimates_smoothing_lambda: float = None,
                       mean_adj_type: ewm.MeanAdjType = ewm.MeanAdjType.NONE
                       ) -> pd.DataFrame:
    """
    ewm convolution, typical case:
    y is return, x is signal
    span deines the
    assumed frequency is daily
    """

    signal_span, _ = da.get_period_days(freq=freq, is_calendar=True)

    if not np.isclose(signal_span, 1):
        ewm_lambda = 1.0 - 2.0 / (signal_span + 1.0)
    else:  # take 1.5 for frequency of business day
        ewm_lambda = 0.5 / 2.5

    if is_ra_return:
        ewm_vol = ewm.compute_ewm_vol(data=returns,
                                         ewm_lambda=ewm_lambda,
                                         annualize=False)
        returns = np.divide(returns, ewm_vol, where=np.isclose(ewm_vol, 0.0)==False)

    # rolling returns by the span
    if not np.isclose(signal_span, 1):
        # norm_factor = np.sqrt(signal_span)
        rolling_returns = returns.rolling(signal_span).sum()
    else:
        rolling_returns = returns

    if signals is not None:
        if signal_agg_type == SignalAggType.LAST_VALUE:
            agg_signal = signals
        elif signal_agg_type == SignalAggType.MEAN:
            agg_signal = signals.rolling(signal_span).mean()
        else:
            raise TypeError(f"unknown {signal_agg_type}")

        agg_signal = agg_signal.reindex(index=rolling_returns.index, method='ffill')
        agg_signal = agg_signal.shift(signal_span)  # shift backrard by the span
    else:
        agg_signal = None

    if convolution_type == ConvolutionType.AUTO_CORR:
        x_data = rolling_returns.shift(signal_span) # shift backward by the span
        y_data = rolling_returns
        cross_xy_type = ewm.CrossXyType.CORR

    elif convolution_type == ConvolutionType.SIGNAL_CORR:
        x_data = agg_signal
        y_data = rolling_returns
        cross_xy_type = ewm.CrossXyType.CORR

    elif convolution_type == ConvolutionType.SIGNAL_BETA:
        x_data = agg_signal
        y_data = rolling_returns
        cross_xy_type = ewm.CrossXyType.BETA

    else:
        raise ValueError(f"{convolution_type} is not implemented")

    # compute ewm cross
    corr = ewm.compute_ewm_cross_xy(x_data=x_data,
                                      y_data=y_data,
                                      ewm_lambda=ewm_lambda,
                                      cross_xy_type=cross_xy_type,
                                      mean_adj_type=mean_adj_type)

    if estimates_smoothing_lambda is not None:
        corr = ewm.compute_ewm(data=corr, ewm_lambda=estimates_smoothing_lambda)

    return corr




import functools
import timeit
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from typing import Union
from enum import Enum

import qis.plots as qp
import qis.models as qm


def get_test_data(n1: int = 1000, n2: int = 3, means: Union[float, np.ndarray] = 0.0) -> pd.DataFrame:
    np_data = np.random.normal(means, 1.0, (n1, n2))
    random_nans = np.random.binomial(size=(n1, n2), n=1, p=0.5)
    np_data = np.where(random_nans, np.nan, np_data)
    df = pd.DataFrame(data=np_data,
                      index=pd.date_range(start='1Jan2020', periods=n1, freq='D'),
                      columns=[f"x{m+1}" for m in range(n2)])
    return df


def time_comp_with_pandas_ewm(span: int = 31) -> None:

    big_df = get_test_data(n1=10000, n2=10000)
    np_data = big_df.to_numpy()
    init_value = np.zeros(np_data.shape[1])

    def pd_ewm():
        big_df.ewm(span=span, adjust=False).mean()

    def np_to_ewm():
        qm.ewm_recursion(a=np_data, span=span, init_value=init_value)

    def pd_to_ewm():
        qm.compute_ewm(data=big_df, span=span, init_value=init_value)

    n = 20
    print(f"pandas ewm: {timeit.Timer(functools.partial(pd_ewm)).timeit(n) / n:.3f}")
    print(f"np_to_ewm: {timeit.Timer(functools.partial(np_to_ewm)).timeit(n) / n:.3f}")
    print(f"pd_to_ewm: {timeit.Timer(functools.partial(pd_to_ewm)).timeit(n) / n:.3f}")


def vol_comp_with_pandas_ewm(span: int = 31) -> None:
    df = get_test_data(n1=10000, n2=1)
    vol_pd = df.ewm(span=span, adjust=False).std()
    vol_ewm1 = qm.compute_ewm_vol(data=df, span=span, mean_adj_type=qm.MeanAdjType.NONE)
    vol_ewm2 = qm.compute_ewm_vol(data=df, span=span, mean_adj_type=qm.MeanAdjType.EWMA)
    joint = pd.concat([vol_pd.iloc[:, 0].rename('Pandas std'),
                      vol_ewm1.iloc[:, 0].rename('Ewm no adj'),
                      vol_ewm2.iloc[:, 0].rename('Ewm with Ewm adj')],
                      axis=1)

    qp.plot_time_series(df=joint,
                        title='Vols comparision',
                        legend_loc='upper left',
                        trend_line=qp.TrendLine.AVERAGE,
                        var_format='{:.2f}')


class UnitTests(Enum):
    TIME_COMP_WITH_PANDAS_EWM = 1
    VOL_COMP_WITH_PANDAS_EWM = 2
    EWMA_NP = 3
    EWMA_DF = 4
    EWMA_MEAN = 5
    EWMA_VOL = 6
    EWMA_BETA = 7
    EWMA_AUTO_CORR = 8
    EWMA_CORR_MATRIX = 9


def run_unit_test(unit_test: UnitTests):

    np.random.seed(1)

    if unit_test == UnitTests.TIME_COMP_WITH_PANDAS_EWM:
        time_comp_with_pandas_ewm()

    elif unit_test == UnitTests.VOL_COMP_WITH_PANDAS_EWM:
        vol_comp_with_pandas_ewm()

    else:

        data = get_test_data(n1=10000, n2=3, means=np.array([-2.0, -1.0, 0.0]))
        ewm_lambda = 0.94
        ewm_lambda3 = np.array([0.94, 0.50, 0.10])
        plot_data, title = None, None

        if unit_test == UnitTests.EWMA_NP:
            ewm1 = qm.ewm_recursion(a=data.iloc[:, 0].to_numpy(dtype=np.double), ewm_lambda=ewm_lambda, init_value=0.0)
            print(ewm1)

            ewm2 = qm.ewm_recursion(a=data.to_numpy(), ewm_lambda=ewm_lambda, init_value=np.zeros(len(data.columns)))
            print(ewm2)

            ewm3 = qm.ewm_recursion(a=data.to_numpy(), ewm_lambda=ewm_lambda3, init_value=np.zeros(len(data.columns)))
            print(ewm3)

        elif unit_test == UnitTests.EWMA_DF:

            ewm_df1 = qm.compute_ewm(data=data.iloc[:, 0], ewm_lambda=ewm_lambda, init_type=qm.InitType.MEAN)
            print(ewm_df1)

            ewm_df1_np = qm.compute_ewm(data=data.iloc[:, 0].to_numpy(), ewm_lambda=ewm_lambda, init_type=qm.InitType.MEAN)
            print(ewm_df1_np)

            ewm_df2_np = qm.compute_ewm(data=data.to_numpy(), ewm_lambda=ewm_lambda, init_type=qm.InitType.MEAN)
            print(ewm_df2_np)

            ewm_df2 = qm.compute_ewm(data=data, ewm_lambda=ewm_lambda, init_type=qm.InitType.MEAN)
            print(ewm_df2)

            ewm_df3 = qm.compute_ewm(data=data, ewm_lambda=ewm_lambda3, init_type=qm.InitType.MEAN)
            print(ewm_df3)

            ewm_df3.columns = [x + ' ewm' for x in data.columns]
            plot_data = pd.concat([data, ewm_df3], axis=1)
            title = 'ewm'

        elif unit_test == UnitTests.EWMA_MEAN:

            datas = {'np1': data.iloc[:, 0].to_numpy(), 'np': data.to_numpy(), 'series': data.iloc[:, 0], 'df': data}

            for key, data_ in datas.items():
                print(key)
                ewm_mean = qm.compute_roll_mean(data=data,
                                                 mean_adj_type=qm.MeanAdjType.EWMA,
                                                 ewm_lambda=ewm_lambda)
                data_adjusted = data - ewm_mean
                ewm_data = qm.compute_ewm(data=data_adjusted,
                                              ewm_lambda=ewm_lambda)

            ewm_mean.columns = [x + ' ewm_mean' for x in data.columns]
            ewm_data.columns = [x + '-ewm_mean' for x in data.columns]
            plot_data = pd.concat([data, ewm_mean, ewm_data], axis=1)
            title = 'ewm-mean'

        elif unit_test == UnitTests.EWMA_VOL:

            datas = {'np1': data.iloc[:, 0].to_numpy(), 'np': data.to_numpy(), 'series': data.iloc[:, 0], 'df': data}

            for key, data_ in datas.items():
                print(key)
                ewm_data = qm.compute_ewm_vol(data=data_,
                                                  ewm_lambda=ewm_lambda,
                                                  mean_adj_type=qm.MeanAdjType.EWMA,
                                                  init_type=qm.InitType.MEAN)
            print(ewm_data)
            ewm_data.columns = [x + ' ewm vol' for x in data.columns]
            data.columns = [x + '^2' for x in data.columns]
            plot_data = pd.concat([np.power(data, 2), ewm_data], axis=1)
            title = 'ewm vol'

        elif unit_test == UnitTests.EWMA_BETA:

            cross_xy_types = [qm.CrossXyType.COVAR, qm.CrossXyType.BETA, qm.CrossXyType.CORR]

            for cross_xy_type in cross_xy_types:
                print(cross_xy_type)
                ewm_data = qm.compute_ewm_cross_xy(x_data=data[[data.columns[0]]],
                                                       y_data=data,
                                                       cross_xy_type=cross_xy_type,
                                                       ewm_lambda=ewm_lambda,
                                                       mean_adj_type=qm.MeanAdjType.EWMA,
                                                       init_type=qm.InitType.MEAN)
            ewm_data.columns = [x + ' ewm beta' for x in data.columns]
            plot_data = pd.concat([data, ewm_data], axis=1)
            title = 'ewm beta'

        elif unit_test == UnitTests.EWMA_AUTO_CORR:
            plot_data = qm.compute_dynamic_auto_corr(data=data,
                                                     ewm_lambda=ewm_lambda,
                                                     mean_adj_type=qm.MeanAdjType.NONE,
                                                     aggregation_type='median',
                                                     is_normalize=True)
            title = 'ewm auto-corr'

        else:
            return

        if plot_data is not None:
            qp.plot_time_series(df=plot_data,
                                title=title,
                                legend_loc='upper left',
                                trend_line=qp.TrendLine.AVERAGE,
                                var_format='{:.2f}')

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.EWMA_NP

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
utiliti
"""
# packages
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from enum import Enum
from dataclasses import dataclass
from typing import NamedTuple, Dict, Union

# qis
from qis.utils.dates import TimePeriod


class TrainLivePeriod(NamedTuple):
    train: TimePeriod
    live: TimePeriod


@dataclass
class TrainLiveSamples:
    train_live_dates: Dict[TimePeriod, TrainLivePeriod] = None

    def __post_init__(self):
        self.train_live_dates = {}

    def add(self, date: TimePeriod, train_live_period: TrainLivePeriod):
        self.train_live_dates[date] = train_live_period

    def print(self):
        for key, samples in self.train_live_dates.items():
            print(f"{key}: train={samples.train.to_str()}, live={samples.live.to_str()}")


def split_to_train_live_samples(ts_index: Union[pd.DatetimeIndex, pd.Index],
                                model_update_freq: 'str' = 'M',
                                roll_period: int = 12
                                ) -> TrainLiveSamples:

    update_dates = pd.date_range(start=ts_index[0],
                                 end=ts_index[-1],
                                 freq=model_update_freq,
                                 closed='right')

    train_live_samples = TrainLiveSamples()
    for idx, date in enumerate(update_dates):
        # x_data is shifted backward
        # last point in y_data is applied for forecast
        # data used for estimation is [last_update, current update date]
        model_update_date = update_dates[idx - 1]

        if idx > roll_period + 1:
            date_t_0 = update_dates[idx-roll_period-1]
            train = TimePeriod(start=date_t_0, end=model_update_date).shift_end_date_by_days(backward=True)
            live = TimePeriod(start=model_update_date, end=date).shift_start_date_by_days(backward=False)

            train_live_samples.add(model_update_date, TrainLivePeriod(train, live))

    return train_live_samples


def split_to_samples(data: Union[pd.DataFrame, pd.Series],
                     sample_freq: 'str' = 'A',
                     start_to_one: bool = False
                     ) -> Dict[pd.Timestamp, pd.DataFrame]:
    data1 = data.resample(sample_freq).last()
    ts_index = data1.index
    update_dates = pd.date_range(start=ts_index[0],
                                 end=ts_index[-1],
                                 freq=sample_freq)
    data_samples = {}
    for idx, date in enumerate(update_dates):
        if idx > 1 and date < ts_index[-1]:
            period_data = data.loc[update_dates[idx-1]: date]
            if start_to_one:
                period_data = period_data.divide(period_data.iloc[0])
            data_samples[date] = period_data

    return data_samples


def get_data_samples_df(data: Union[pd.DataFrame, pd.Series],
                        sample_freq: 'str' = 'A',
                        start_to_one: bool = False
                        ) -> pd.DataFrame:
    data_samples = {}
    data_samples_dict = split_to_samples(data, sample_freq=sample_freq, start_to_one=start_to_one)
    for key, kdata in data_samples_dict.items():
        data_samples[key] = kdata.reset_index(drop=True)
    data_samples_df = pd.DataFrame.from_dict(data_samples)
    data_samples_df = data_samples_df.fillna(method='ffill')
    return data_samples_df


class UnitTests(Enum):
    SAMPLE_DATES = 1
    SPLIT_TO_SAMPLES = 2


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.SAMPLE_DATES:
        time_period = TimePeriod(start='31Dec2018', end='31Dec2020')

        ts_index = time_period.to_pd_datetime_index(freq='M')
        train_live_samples = split_to_train_live_samples(ts_index=ts_index, model_update_freq='M', roll_period=12)
        train_live_samples.print()

    elif unit_test == UnitTests.SPLIT_TO_SAMPLES:
        time_period = TimePeriod(start='31Dec2010', end='31Dec2020')

        ts_index = time_period.to_pd_datetime_index(freq='B')
        data = pd.DataFrame(data=np.random.normal(0, 1.0, (len(ts_index), 1)), index=ts_index, columns=['id1'])

        data_samples = split_to_samples(data=data, sample_freq='A')
        print(data_samples)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.SPLIT_TO_SAMPLES

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)




from qis.portfolio.backtester import backtest_model_portfolio

from qis.portfolio.portfolio_data import PortfolioData, PortfolioInput

from qis.portfolio.multi_portfolio_data import MultiPortfolioData

from qis.portfolio.reports.multi_strategy_factsheet import generate_multi_portfolio_factsheet




# buil in
import numpy as np
import pandas as pd
from enum import Enum
from typing import NamedTuple


class OhlcEstimatorType(Enum):
    PARKINSON = 'Parkinson'
    GARMAN_KLASS = 'Garman-Klass'
    ROGERS_SATCHELL = 'Rogers-Satchell'


class FreqAn(NamedTuple):
    freq: str
    an: float


class FreqAns(FreqAn, Enum):
    HOUR = FreqAn('1h', an=24*260)
    DAY = FreqAn('D', an=260)


def estimate_ohlc_vol(ohlc_data: pd.DataFrame,  # must contain ohlc columnes
                      ohlc_estimator_type: OhlcEstimatorType = OhlcEstimatorType.PARKINSON,
                      min_size: int = 2
                      ) -> float:

    if ohlc_data.empty or len(ohlc_data.index) < min_size:
        return np.nan

    log_ohlc = np.log(ohlc_data[['open', 'high', 'low', 'close']].to_numpy())
    open, high, low, close = log_ohlc[:, 0], log_ohlc[:, 1], log_ohlc[:, 2], log_ohlc[:, 3]

    hc = high - close
    ho = high - open
    lc = low - close
    lo = low - open
    hl = high - low
    co = close - open

    if ohlc_estimator_type == OhlcEstimatorType.PARKINSON:
        multiplier = 1.0 / (4.0 * np.log(2.0))
        sample_var = multiplier * np.square(hl)

    elif ohlc_estimator_type == OhlcEstimatorType.GARMAN_KLASS:
        multiplier = 2.0 * np.log(2.0) - 1.0
        sample_var = 0.5 * np.square(hl) - multiplier * np.square(co)

    elif ohlc_estimator_type == OhlcEstimatorType.ROGERS_SATCHELL:
        sample_var = hc*ho + lc*lo

    else:
        raise TypeError(f"unknown ohlc_estimator_type={ohlc_estimator_type}")

    vol = np.sqrt(np.nansum(sample_var))
    return vol


def estimate_intra_ohlc_vol_data(ohlc_data: pd.DataFrame,
                                 ohlc_estimator_type: OhlcEstimatorType = OhlcEstimatorType.PARKINSON,
                                 is_exclude_weekends: bool = True,
                                 freq_an: FreqAn = FreqAns.DAY
                                 ) -> pd.Series:
    """
    sample vol at freq and annualize at an
    """
    vols = ohlc_data.groupby(pd.Grouper(freq=freq_an.freq)).apply(estimate_ohlc_vol, ohlc_estimator_type)
    vols = np.sqrt(freq_an.an) * vols  # annualize
    if is_exclude_weekends:
        vols = vols[vols.index.dayofweek < 5]
    return vols



"""
generate portfolio factsheet report using PortfolioData data object
with comparision to 1-2 cash benchmarks
"""
# packages
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from typing import Tuple
from enum import Enum

# qis
import qis
import qis.utils as qu
import qis.plots as qp
from qis.utils.dates import TimePeriod
from qis.perfstats.config import PerfParams
from qis.perfstats.regime_classifier import BenchmarkReturnsQuantileRegimeSpecs

# portfolio
import qis.portfolio.backtester as bp
from qis.portfolio.portfolio_data import PortfolioData

PERF_PARAMS = PerfParams(freq='W-WED')
REGIME_PARAMS = BenchmarkReturnsQuantileRegimeSpecs(freq='Q')

# use for number years > 5
KWARG_LONG = dict(perf_params=PerfParams(freq='W-WED', freq_reg='Q'),
                  regime_params=BenchmarkReturnsQuantileRegimeSpecs(freq='Q'),
                  x_date_freq='A',
                  date_format='%b-%y')

# use for number years < 3
KWARG_SHORT = dict(perf_params=PerfParams(freq='W-WED', freq_reg='M'),
                   regime_params=BenchmarkReturnsQuantileRegimeSpecs(freq='M'),
                   x_date_freq='Q',
                   date_format='%b-%y')


def generate_portfolio_factsheet(portfolio_data: PortfolioData,
                                 benchmark_prices: pd.DataFrame,
                                 time_period: TimePeriod = None,
                                 perf_params: PerfParams = PERF_PARAMS,
                                 regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                                 figsize: Tuple[float, float] = (8.3, 11.7),  # A4 for portrait
                                 **kwargs
                                 ) -> plt.Figure:
    # align
    benchmark_prices = benchmark_prices.reindex(index=portfolio_data.nav.index, method='ffill')
    regime_benchmark = benchmark_prices.columns[0]

    fig = plt.figure(figsize=figsize, constrained_layout=True)
    gs = fig.add_gridspec(nrows=14, ncols=4, wspace=0.0, hspace=0.0)

    plot_kwargs = dict(fontsize=5,
                       linewidth=0.5,
                       digits_to_show=1, sharpe_digits=2,
                       weight='normal',
                       markersize=1,
                       framealpha=0.75)
    kwargs = qu.update_kwargs(kwargs, plot_kwargs)
    fig.suptitle(f"{portfolio_data.nav.name} portfolio factsheet",
                 fontweight="bold", fontsize=8, color='blue')

    # prices
    joint_prices = pd.concat([portfolio_data.get_portfolio_nav(time_period=time_period), benchmark_prices], axis=1).dropna()
    pivot_prices = joint_prices[regime_benchmark]
    ax = fig.add_subplot(gs[0:2, :2])
    qp.plot_prices(prices=joint_prices,
                   perf_params=perf_params,
                   title='Performance',
                   ax=ax,
                   **kwargs)
    qp.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
    qp.set_spines(ax=ax, bottom_spine=False, left_spine=False)

    # dd
    ax = fig.add_subplot(gs[2:4, :2])
    qp.plot_drawdown(prices=joint_prices,
                      title='Running Drawdowns',
                      dd_legend_type=qp.DdLegendType.SIMPLE,
                      ax=ax, **kwargs)
    qp.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
    qp.set_spines(ax=ax, bottom_spine=False, left_spine=False)

    # exposures
    if len(portfolio_data.weights.columns) > 10:
        exposures = portfolio_data.get_exposures(is_grouped=True, time_period=time_period)
    else:
        exposures = portfolio_data.get_exposures(is_grouped=False, time_period=time_period)
    ax = fig.add_subplot(gs[4:6, :2])
    qp.plot_stack(df=exposures.resample('W-WED').last(),
                  add_mean_levels=False,
                  is_use_bar_plot=True,
                  baseline='zero',
                  title='Exposures',
                  legend_stats=qp.LegendStats.AVG_LAST,
                  var_format='{:.1%}',
                  ax=ax,
                  **qu.update_kwargs(kwargs, dict(bbox_to_anchor=(0.5, 1.05), ncol=2)))
    qp.set_spines(ax=ax, bottom_spine=False, left_spine=False)

    # turnover
    ax = fig.add_subplot(gs[6:8, :2])
    turnover = portfolio_data.get_turnover(time_period=time_period, roll_period=260)

    qp.plot_time_series(df=turnover,
                         var_format='{:,.2%}',
                         # y_limits=(0.0, None),
                         legend_stats=qp.LegendStats.AVG_LAST,
                         title='1y rolling average Turnover',
                         ax=ax,
                         **kwargs)
    """
    qp.plot_time_series_2ax(df1=turnover.drop(portfolio_data.nav.name, axis=1),
                             df2=turnover[portfolio_data.nav.name],
                             var_format='{:,.2%}',
                             # y_limits=(0.0, None),
                             legend_stats=pts.LegendStats.AVG_LAST,
                             title='1y rolling average Turnover',
                             ax=ax,
                             **kwargs)
    """
    qp.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
    qp.set_spines(ax=ax, bottom_spine=False, left_spine=False)
    
    # benchmark betas
    ax = fig.add_subplot(gs[8:10, :2])
    factor_exposures = portfolio_data.compute_portfolio_benchmark_betas(benchmark_prices=benchmark_prices,
                                                                        time_period=time_period)
    qp.plot_time_series(df=factor_exposures,
                         var_format='{:,.2f}',
                         # y_limits=(0.0, None),
                         legend_stats=qp.LegendStats.AVG_LAST,
                         title='Portfolio Benchmark betas',
                         ax=ax,
                         **kwargs)
    qp.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
    qp.set_spines(ax=ax, bottom_spine=False, left_spine=False)

    # attribution
    ax = fig.add_subplot(gs[10:12, :2])
    factor_attribution = portfolio_data.compute_portfolio_benchmark_attribution(benchmark_prices=benchmark_prices,
                                                                                time_period=time_period)
    qp.plot_time_series(df=factor_attribution,
                         var_format='{:,.0%}',
                         # y_limits=(0.0, None),
                         legend_stats=qp.LegendStats.LAST,
                         title='Portfolio Cumulative return attribution to benchmark betas',
                         ax=ax,
                         **kwargs)
    qp.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
    qp.set_spines(ax=ax, bottom_spine=False, left_spine=False)

    # constituents
    ax = fig.add_subplot(gs[12:, :2])
    num_investable_instruments=portfolio_data.get_num_investable_instruments(time_period=time_period)
    qp.plot_time_series(df=num_investable_instruments,
                         var_format='{:,.0f}',
                         legend_stats=qp.LegendStats.FIRST_AVG_LAST,
                         title='Number of investable instruments',
                         ax=ax,
                         **kwargs)
    qp.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
    qp.set_spines(ax=ax, bottom_spine=False, left_spine=False)

    # ra perf table
    ax = fig.add_subplot(gs[0, 2:])
    portfolio_data.plot_ra_perf_table(ax=ax,
                                      benchmark_price=benchmark_prices.iloc[:, 0],
                                      time_period=time_period,
                                      perf_params=perf_params,
                                      **qu.update_kwargs(kwargs, dict(fontsize=4)))
    ax = fig.add_subplot(gs[1, 2:])
    portfolio_data.plot_ra_perf_table(ax=ax,
                                      benchmark_price=benchmark_prices.iloc[:, 0],
                                      time_period=qu.get_time_period_shifted_by_years(time_period=time_period),
                                      perf_params=perf_params,
                                      **qu.update_kwargs(kwargs, dict(fontsize=4)))

    # heatmap
    ax = fig.add_subplot(gs[2:4, 2:])
    portfolio_data.plot_monthly_returns_heatmap(ax=ax,
                                                time_period=time_period,
                                                title='Monthly Returns',
                                                **qu.update_kwargs(kwargs, dict(fontsize=4)))

    # periodic returns
    ax = fig.add_subplot(gs[4:6, 2:])
    local_kwargs = qu.update_kwargs(kwargs=kwargs,
                                     new_kwargs=dict(fontsize=4, square=False, x_rotation=90, transpose=True))
    portfolio_data.plot_periodic_returns(ax=ax,
                                         benchmark_prices=benchmark_prices,
                                         heatmap_freq='A',
                                         time_period=time_period,
                                         **qu.update_kwargs(local_kwargs, dict(date_format='%Y')))

    # perf contributors
    ax = fig.add_subplot(gs[6:8, 2])
    portfolio_data.plot_contributors(ax=ax,
                                     time_period=time_period,
                                     title=f"Performance Contributors {time_period.to_str()}",
                                     **kwargs)

    ax = fig.add_subplot(gs[6:8, 3])
    time_period_1y = qu.get_time_period_shifted_by_years(time_period=time_period)
    portfolio_data.plot_contributors(ax=ax,
                                     time_period=time_period_1y,
                                     title=f"Performance Contributors {time_period_1y.to_str()}",
                                     **kwargs)

    # regime data
    ax = fig.add_subplot(gs[8:10, 2:])
    portfolio_data.plot_regime_data(ax=ax,
                                    benchmark_price=benchmark_prices.iloc[:, 0],
                                    time_period=time_period,
                                    perf_params=perf_params,
                                    regime_params=regime_params,
                                    **kwargs)

    # vol regime data
    """
    ax = fig.add_subplot(gs[10:12, 2:])
    portfolio_data.plot_vol_regimes(ax=ax,
                                    benchmark_price=benchmark_prices.iloc[:, 0],
                                    time_period=time_period,
                                    perf_params=perf_params,
                                    regime_params=regime_params,
                                    **kwargs)
    """
    # returns scatter
    ax = fig.add_subplot(gs[10:12, 2:])
    portfolio_data.plot_returns_scatter(ax=ax,
                                        benchmark_price=benchmark_prices.iloc[:, 0],
                                        time_period=time_period,
                                        freq=perf_params.freq_reg,
                                        **kwargs)

    if len(benchmark_prices.columns) > 1:
        ax = fig.add_subplot(gs[12:, 2:])
        portfolio_data.plot_returns_scatter(ax=ax,
                                            benchmark_price=benchmark_prices.iloc[:, 1],
                                            time_period=time_period,
                                            **kwargs)

    return fig


class UnitTests(Enum):
    TEST1 = 1
    TEST2 = 2


def run_unit_test(unit_test: UnitTests):

    from qis.test_data import load_etf_data
    prices = load_etf_data()#.dropna()
    time_period = TimePeriod('31Dec2001', '31Dec2022')
    prices = time_period.locate(prices)
    group_data = pd.Series(dict(SPY='Equities', QQQ='Equities', EEM='Equities', TLT='Bonds', IEF='Bonds', SHY='Bonds', LQD='Credit', HYG='HighYield', GLD='Gold'))

    benchmark_prices = prices[['SPY', 'TLT']]

    portfolio_data = bp.backtest_model_portfolio(prices=prices,
                                                 weights=np.ones(len(prices.columns)) / len(prices.columns),
                                                 is_output_portfolio_data=True,
                                                 ticker='EqualWeighted')
    portfolio_data._set_group_data(group_data=group_data, group_order=list(group_data.unique()))

    if unit_test == UnitTests.TEST1:
        fig = generate_portfolio_factsheet(portfolio_data=portfolio_data,
                                     benchmark_prices=benchmark_prices,
                                     time_period=TimePeriod('31Dec2005', '31Dec2022'),
                                     **KWARG_LONG)
        qis.save_figs_to_pdf(figs=[fig],
                             file_name=f"{portfolio_data.nav.name}_portfolio_factsheet",
                             orientation='landscape',
                             local_path=qis.local_path.get_output_path())

    if unit_test == UnitTests.TEST2:
        fig = generate_portfolio_factsheet(portfolio_data=portfolio_data,
                                           benchmark_prices=benchmark_prices,
                                           time_period=TimePeriod('31Dec2019', '31Dec2022'),
                                           **KWARG_SHORT)
        qis.save_figs_to_pdf(figs=[fig],
                             file_name=f"{portfolio_data.nav.name}_portfolio_factsheet",
                             orientation='landscape',
                             local_path=qis.local_path.get_output_path())

        plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.TEST1

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
generate strategy factsheet report using MultiPortfolioData data object
with comparision to 1-2 cash benchmarks
"""
# packages
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from typing import List, Tuple
import seaborn as sns
from enum import Enum

# qis
import qis
import qis.file_utils as fu
import qis.utils as qu
from qis.perfstats.config import PerfParams
from qis.perfstats.regime_classifier import BenchmarkReturnsQuantileRegimeSpecs

# portfolio
import qis.portfolio.backtester as bp
from qis.portfolio.portfolio_data import AttributionMetric
from qis.portfolio.multi_portfolio_data import MultiPortfolioData
from qis.portfolio.reports.portfolio_factsheet import generate_portfolio_factsheet

PERF_PARAMS = PerfParams(freq='W-WED')
REGIME_PARAMS = BenchmarkReturnsQuantileRegimeSpecs(freq='Q')


# use for number years > 5
KWARG_LONG = dict(perf_params=PerfParams(freq='W-WED', freq_reg='Q'),
                  regime_params=BenchmarkReturnsQuantileRegimeSpecs(freq='Q'),
                  x_date_freq='A',
                  date_format='%b-%y')

# use for number years < 3
KWARG_SHORT = dict(perf_params=PerfParams(freq='W-WED', freq_reg='M'),
                   regime_params=BenchmarkReturnsQuantileRegimeSpecs(freq='M'),
                   x_date_freq='Q',
                   date_format='%b-%y')


def generate_strategy_factsheet(multi_portfolio_data: MultiPortfolioData,
                                time_period: qu.TimePeriod = None,
                                perf_params: PerfParams = PERF_PARAMS,
                                regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                                backtest_name: str = None,
                                add_strategy_factsheets: bool = False,
                                figsize: Tuple[float, float] = (8.3, 11.7),  # A4 for portrait
                                **kwargs
                                ) -> plt.Figure:
    """
    designed for 1 strategy and 1 benchmark report
    multi_portfolio_data = [stragegy portfolio, benchmark strategy portfolio]
    """
    if len(multi_portfolio_data.portfolio_datas) == 1:
        raise ValueError(f"must be at least two strategieg")

    plot_kwargs = dict(fontsize=5,
                       linewidth=0.5,
                       digits_to_show=1, sharpe_digits=2,
                       weight='normal',
                       markersize=1,
                       framealpha=0.75,
                       time_period=time_period)

    kwargs = qu.update_kwargs(kwargs, plot_kwargs)
    fig = plt.figure(figsize=figsize, constrained_layout=True)
    gs = fig.add_gridspec(nrows=14, ncols=4, wspace=0.0, hspace=0.0)

    if backtest_name is not None:
        fig.suptitle(backtest_name, fontweight="bold", fontsize=8, color='blue')

    regime_benchmark = multi_portfolio_data.benchmark_prices.columns[0]
    benchmark_price = multi_portfolio_data.benchmark_prices[regime_benchmark]

    multi_portfolio_data.plot_nav(ax=fig.add_subplot(gs[:2, :2]),
                                  benchmark=regime_benchmark,
                                  perf_params=perf_params,
                                  regime_params=regime_params,
                                  title='Cumulative performance',
                                  **kwargs)

    multi_portfolio_data.plot_drawdowns(ax=fig.add_subplot(gs[2:4, :2]),
                                        benchmark=regime_benchmark,
                                        regime_params=regime_params,
                                        title='Running Drawdowns',
                                        **kwargs)

    multi_portfolio_data.plot_rolling_time_under_water(ax=fig.add_subplot(gs[4:6, :2]),
                                                       benchmark=regime_benchmark,
                                                       regime_params=regime_params,
                                                       title='Rolling time under water',
                                                       **kwargs)

    multi_portfolio_data.plot_exposures(ax=fig.add_subplot(gs[6:8, :2]),
                                        benchmark=regime_benchmark,
                                        regime_params=regime_params,
                                        **kwargs)

    multi_portfolio_data.plot_exposures_diff(ax=fig.add_subplot(gs[8:10, :2]),
                                             benchmark=regime_benchmark,
                                             regime_params=regime_params,
                                             **kwargs)

    multi_portfolio_data.plot_turnover(ax=fig.add_subplot(gs[10:12, :2]),
                                       benchmark=regime_benchmark,
                                       regime_params=regime_params,
                                       **kwargs)

    multi_portfolio_data.plot_costs(ax=fig.add_subplot(gs[12:14, :2]),
                                    benchmark=regime_benchmark,
                                    regime_params=regime_params,
                                    **kwargs)

    multi_portfolio_data.plot_ac_ra_perf_table(ax=fig.add_subplot(gs[0:2, 2:]),
                                               benchmark_price=benchmark_price,
                                               perf_params=perf_params,
                                               **qu.update_kwargs(kwargs, dict(fontsize=4)))

    time_period1 = qu.get_time_period_shifted_by_years(time_period=time_period)
    multi_portfolio_data.plot_ac_ra_perf_table(ax=fig.add_subplot(gs[2:4, 2:]),
                                               benchmark_price=benchmark_price,
                                               perf_params=perf_params,
                                               **qu.update_kwargs(kwargs, dict(time_period=time_period1, fontsize=4)))

    # periodic returns

    local_kwargs = qu.update_kwargs(kwargs=kwargs,
                                    new_kwargs=dict(fontsize=4, square=False, x_rotation=90, transpose=False))
    multi_portfolio_data.portfolio_datas[0].plot_periodic_returns(ax=fig.add_subplot(gs[4:6, 2]),
                                                                  heatmap_freq='A',
                                                                  **qu.update_kwargs(local_kwargs, dict(date_format='%Y')))

    multi_portfolio_data.portfolio_datas[1].plot_periodic_returns(ax=fig.add_subplot(gs[4:6, 3]),
                                                                  heatmap_freq='A',
                                                                  **qu.update_kwargs(local_kwargs, dict(date_format='%Y')))

    multi_portfolio_data.portfolio_datas[0].plot_regime_data(ax=fig.add_subplot(gs[6:8, 2]),
                                                             benchmark_price=benchmark_price,
                                                             title=f"{multi_portfolio_data.portfolio_datas[0].nav.name}",
                                                             perf_params=perf_params,
                                                             regime_params=regime_params,
                                                             **qu.update_kwargs(kwargs, dict(fontsize=4, x_rotation=90)))

    multi_portfolio_data.portfolio_datas[1].plot_regime_data(ax=fig.add_subplot(gs[6:8, 3]),
                                                             benchmark_price=benchmark_price,
                                                             title=f"{multi_portfolio_data.portfolio_datas[1].nav.name}",
                                                             perf_params=perf_params,
                                                             regime_params=regime_params,
                                                             **qu.update_kwargs(kwargs, dict(fontsize=4, x_rotation=90)))

    # vol regimes
    """
    multi_portfolio_data.portfolio_datas[0].plot_vol_regimes(ax=fig.add_subplot(gs[8:10, 2]),
                                                             benchmark_price=benchmark_price,
                                                             perf_params=perf_params,
                                                             regime_params=regime_params,
                                                             **qu.update_kwargs(kwargs, dict(fontsize=4, x_rotation=90)))
    multi_portfolio_data.portfolio_datas[1].plot_vol_regimes(ax=fig.add_subplot(gs[8:10, 3]),
                                                             benchmark_price=benchmark_price,
                                                             perf_params=perf_params,
                                                             regime_params=regime_params,
                                                             **qu.update_kwargs(kwargs, dict(fontsize=4, x_rotation=90)))
    """
    multi_portfolio_data.plot_instrument_pnl_diff(ax=fig.add_subplot(gs[8:10, 2:]),
                                                  benchmark=regime_benchmark,
                                                  regime_params=regime_params,
                                                  **kwargs)

    multi_portfolio_data.plot_factor_betas(ax=fig.add_subplot(gs[10:12, 2:]),
                                           benchmark_prices=multi_portfolio_data.benchmark_prices,
                                           benchmark=regime_benchmark,
                                           regime_params=regime_params,
                                           **kwargs)

    multi_portfolio_data.plot_returns_scatter(ax=fig.add_subplot(gs[12:, 2:]),
                                              benchmark=regime_benchmark,
                                              **kwargs)

    figs = [fig]
    if add_strategy_factsheets:
        for portfolio_data in multi_portfolio_data.portfolio_datas:
            figs.append(generate_portfolio_factsheet(portfolio_data=portfolio_data,
                                                     benchmark_prices=multi_portfolio_data.benchmark_prices,
                                                     time_period=time_period,
                                                     perf_params=perf_params,
                                                     regime_params=regime_params))

    return fig


def generate_performance_attribution_report(multi_portfolio_data: MultiPortfolioData,
                                            time_period: qu.TimePeriod = None,
                                            figsize: Tuple[float, float] = (8.3, 11.7),  # A4 for portrait
                                            **kwargs
                                            ) -> List[plt.Figure]:
    figs = []
    with sns.axes_style('darkgrid'):
        fig = plt.figure(figsize=figsize, constrained_layout=True)
        figs.append(fig)
        gs = fig.add_gridspec(nrows=3, ncols=2, wspace=0.0, hspace=0.0)

        multi_portfolio_data.plot_nav(ax=fig.add_subplot(gs[0, :]),
                                      time_period=time_period,
                                      **kwargs)

        multi_portfolio_data.plot_performance_attribution(portfolio_ids=[0],
                                                          time_period=time_period,
                                                          attribution_metric=AttributionMetric.PNL,
                                                          ax=fig.add_subplot(gs[1, :]),
                                                          **kwargs)

        multi_portfolio_data.plot_performance_attribution(portfolio_ids=[1],
                                                          time_period=time_period,
                                                          attribution_metric=AttributionMetric.PNL_RISK,
                                                          ax=fig.add_subplot(gs[2, :]),
                                                          **kwargs)

        fig, ax = plt.subplots(1, 1, figsize=figsize)
        multi_portfolio_data.plot_performance_attribution(portfolio_ids=[0],
                                                          time_period=time_period,
                                                          attribution_metric=AttributionMetric.PNL_RISK,
                                                          ax=ax,
                                                          **kwargs)

        fig, ax = plt.subplots(1, 1, figsize=figsize)
        figs.append(fig)
        multi_portfolio_data.plot_performance_periodic_table(portfolio_ids=[0],
                                                             time_period=time_period,
                                                             attribution_metric=AttributionMetric.INST_PNL,
                                                             freq='M',
                                                             ax=ax,
                                                             **kwargs)

        fig, ax = plt.subplots(1, 1, figsize=figsize)
        figs.append(fig)
        multi_portfolio_data.plot_composite_table(ax=ax, **kwargs)

    return figs


class UnitTests(Enum):
    PERF_ATTRIBUTION = 1
    FACTSHEET = 2


def run_unit_test(unit_test: UnitTests):
    from qis.test_data import load_etf_data
    import qis.models.linear.ra_returns as rar

    prices = load_etf_data()  # .dropna()
    ra_returns, weights, ewm_vol = rar.compute_ra_returns(returns=prices.pct_change(), span=60, vol_target=0.15)
    weights = weights.divide(weights.sum(1), axis=0)

    time_period = qu.TimePeriod('31Dec2003', '31Dec2022')
    prices = time_period.locate(prices)
    group_data = pd.Series(
        dict(SPY='Equities', QQQ='Equities', EEM='Equities', TLT='Bonds', IEF='Bonds', SHY='Bonds', LQD='Credit',
             HYG='HighYield', GLD='Gold'))

    benchmark_prices = prices[['SPY', 'TLT']]

    portfolio_data1 = bp.backtest_model_portfolio(prices=prices,
                                                  weights=time_period.locate(weights),
                                                  is_output_portfolio_data=True,
                                                  ticker='VolParity')
    portfolio_data1._set_group_data(group_data=group_data, group_order=list(group_data.unique()))

    portfolio_data2 = bp.backtest_model_portfolio(prices=prices,
                                                  weights=np.ones(len(prices.columns)) / len(prices.columns),
                                                  is_output_portfolio_data=True,
                                                  ticker='EW')
    portfolio_data2._set_group_data(group_data=group_data, group_order=list(group_data.unique()))

    multi_portfolio_data = MultiPortfolioData(portfolio_datas=[portfolio_data1, portfolio_data2],
                                              benchmark_prices=benchmark_prices)

    if unit_test == UnitTests.PERF_ATTRIBUTION:

        figs = generate_performance_attribution_report(multi_portfolio_data=multi_portfolio_data,
                                                       time_period=qu.TimePeriod('31Dec2021', '31Dec2022'),
                                                       **KWARG_SHORT)
        fu.save_figs_to_pdf(figs=figs,
                            file_name=f"perf_attribution",
                            orientation='landscape',
                            local_path=qis.local_path.get_output_path())

    if unit_test == UnitTests.FACTSHEET:
        fig = generate_strategy_factsheet(multi_portfolio_data=multi_portfolio_data,
                                          backtest_name='Vol Parity Portfolio vs Equal Weight',
                                          time_period=qu.TimePeriod('31Dec2006', '31Dec2022'),
                                          **KWARG_LONG)
        qis.save_figs_to_pdf(figs=[fig],
                             file_name=f"strategy_factsheet", orientation='landscape',
                             local_path=qis.local_path.get_output_path())
    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.FACTSHEET

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



# imports
import matplotlib.pyplot as plt
import seaborn as sns
import yfinance as yf
import qis

# skip
import qis.file_utils as fu

# define tickers and fetch price data
tickers = ['SPY', 'QQQ', 'EEM', 'TLT', 'IEF', 'LQD', 'HYG', 'GLD']
prices = yf.download(tickers, start=None, end=None)['Adj Close'][tickers].dropna()

# minimum usage
fig = qis.plot_prices(prices=prices)

# skip
fu.save_fig(fig, file_name='perf1', local_path="figures/")

# with drawdowns using sns styles
with sns.axes_style("darkgrid"):
    fig, axs = plt.subplots(2, 1, figsize=(10, 7))
    qis.plot_prices_with_dd(prices=prices, axs=axs)

# skip
fu.save_fig(fig, file_name='perf2', local_path="figures/")

# risk-adjusted performance table with specified data entries
# add rates for excess Sharpe
from qis import PerfStat
ust_3m_rate = yf.download('^IRX', start=None, end=None)['Adj Close'].dropna() / 100.0

# set parameters for computing performance stats including returns vols and regressions
perf_params = qis.PerfParams(freq='M', freq_reg='Q', rates_data=ust_3m_rate)

fig = qis.plot_ra_perf_table(prices=prices,
                             perf_columns=[PerfStat.TOTAL_RETURN, PerfStat.PA_RETURN, PerfStat.VOL, PerfStat.SHARPE,
                                           PerfStat.SHARPE_EXCESS, PerfStat.MAX_DD, PerfStat.MAX_DD_VOL,
                                           PerfStat.SKEWNESS, PerfStat.KURTOSIS],
                             title=f"Risk-adjusted performance: {qis.get_time_period_label(prices, date_separator='-')}",
                             perf_params=perf_params)

# skip
fu.save_fig(fig, file_name='perf3', local_path="figures/")


plt.show()



"""
plot df as table
"""
# packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.table import Table as Table
from enum import Enum
from typing import List, Tuple, Optional, Literal

# qis
import qis.plots.utils as put
import qis.utils.df_str as dfs

ROW_HIGHT = 0.625  # cm
COLUMN_WIDTH = 2.0  # cm ?
FIRST_COLUMN_WIDTH = 3.0


def plot_df_table(df: pd.DataFrame,
                  add_index_as_column: bool = True,
                  column_width: float = COLUMN_WIDTH,
                  first_column_width: Optional[float] = FIRST_COLUMN_WIDTH,
                  row_height: float = ROW_HIGHT,
                  first_row_height: float = None,
                  col_widths: List[float] = None,  # can pass as cols
                  rotation_for_columns_headers: int = None,
                  rotation_for_text: int = None,
                  transpose: bool = False,
                  index_column_name: str = ' ',
                  fontsize: int = 10,
                  header_color: str = '#40466e',
                  header_text_color: str = 'w',
                  row_colors: List[str] = ('#f1f1f2', 'w'),
                  edge_color: str = 'w',
                  bbox: Tuple[float] = (0, 0, 1, 1),
                  header_column_id: int = 0,
                  header_row_id: int = 0,
                  left_aligned_first_col: bool = False,
                  var_format: str = None,  # '{:.2f}' to convert numerical data to str
                  title: str = None,
                  heatmap_columns: List[int] = None,
                  heatmap_rows: List[int] = None,
                  heatmap_rows_columns: Tuple[Tuple[int, int], Tuple[int, int]] = None,  # row[0]: row[1], column[0]:column[1]
                  cmap: str = 'RdYlGn',
                  special_rows_colors: List[Tuple[int, str]] = None,
                  special_columns_colors: List[Tuple[int, str]] = None,
                  data_colors: List[Tuple[float, float, float]] = None,
                  diagonal_color: str = None,
                  rows_edge_lines: List[int] = None,
                  columns_edge_lines: List[Tuple[int, str]] = None,
                  bold_font: bool = False,
                  alpha: float = 1.0,
                  emply_column_names: bool = False,
                  ax: plt.Subplot = None,
                  **kwargs
                  ) -> Optional[plt.Figure]:
    """
    plot dataframe as maotplotlib table
    """
    df = df.copy()  # data object will be changed
    if transpose:
        t_data = df.T
        if add_index_as_column:
            t_data.columns = t_data.iloc[0, :]  # rename columns by original index
            df = t_data.drop(labels=t_data.index[0], axis=0)
        else:
            t_data.columns = df.index  # rename columns using original index
            df = t_data

    if add_index_as_column:
        df.insert(0, column=index_column_name, value=df.index)

    if col_widths is None:
        col_widths = [column_width for _ in df]
    else:
        first_column_width = col_widths[0]

    # allocate size
    size = (np.array(df.shape[::-1]) + np.array([0, 1])) * np.array([col_widths[0], row_height])

    if first_column_width is None:
        first_column_width = col_widths[0]
    #else:
    #    col_widths[0] = first_column_width # after this change

    if ax is None:  # create new axis
        fig, ax = plt.subplots(figsize=size)
        ax.axis('off')
    else:  # add table to existing axis
        fig = None
        ax.axis('off')

    if var_format is not None:
        df = dfs.df_to_str(df=df, var_format=var_format)

    if emply_column_names is False:
        col_labels = df.columns.to_list()
    else:
        col_labels = [df.columns[0]] + [''] * (len(df.columns) - 1)

    mpl_table = ax.table(cellText=df.to_numpy(),
                         bbox=bbox,
                         colLabels=col_labels,
                         colLoc='center')

    mpl_table.auto_set_font_size(False)
    mpl_table.set_fontsize(fontsize)

    weight = 'bold' if bold_font else 'normal'
    for k, cell in mpl_table._cells.items():
        if k[0] == header_row_id or k[1] < header_column_id:
            cell.set_edgecolor(edge_color)
            cell.set_text_props(weight=weight, color=header_text_color)
            cell.set_facecolor(header_color)
            if first_row_height is not None:
                cell.set_height(first_row_height)

            if rotation_for_columns_headers is not None:
                cell.get_text().set_rotation(rotation_for_columns_headers)  # set rotation
                # cell.set_height(first_column_width / 15)  # aling widths

        elif k[0] > header_row_id:
            cell.set_edgecolor(edge_color)
            cell.set_facecolor(row_colors[k[0] % len(row_colors) ])
            cell.set_edgecolor('slategray')

        if k[1] == 0:
            width = first_column_width
        else:
            width = col_widths[k[1]]
        cell.set_width(width)
        if rotation_for_text is not None:
            cell.get_text().set_rotation(rotation_for_text)  # set rotation

    if heatmap_columns is not None:
        for heatmap_column in heatmap_columns:
            column_data = df[df.columns[heatmap_column]]
            x_array = dfs.series_to_numeric(ds=column_data)
            colors = put.compute_heatmap_colors(a=x_array, cmap=cmap)
            for k, cell in mpl_table._cells.items():
                if k[0] > 0 and k[1] == heatmap_column:  # skip first row
                    cell.set_facecolor(colors[k[0]-1])

    if heatmap_rows is not None:
        for heatmap_row in heatmap_rows:
            row_data = df.iloc[heatmap_row, 1:]  # exclude first row
            x_array = dfs.series_to_numeric(ds=row_data)
            colors = put.compute_heatmap_colors(a=x_array, cmap=cmap)
            for k, cell in mpl_table._cells.items():
                if k[1] > 0 and k[0] == heatmap_row+1:  # heatmap_row is not counting first headers
                    cell.set_facecolor(colors[k[1]-1])

    if heatmap_rows_columns is not None:   # row[0]: row[1], column[0]:column[1]
        # set colors on all data
        row_start, row_end = heatmap_rows_columns[0][0], heatmap_rows_columns[0][1]
        col_start, col_end = heatmap_rows_columns[1][0], heatmap_rows_columns[1][1]

        if add_index_as_column:
            col_start = col_start + 1
            col_end = col_end + 1
        else:
            col_start = col_start - 1

        data_extract = df.iloc[row_start: row_end, col_start: col_end]
        a = dfs.df_to_numeric(df=data_extract)
        colors = put.compute_heatmap_colors(a=a, cmap=cmap)
        for row_idx in range(len(data_extract.index)):
            col_idx = 0
            for k, cell in mpl_table._cells.items():
                if k[0] == row_start + row_idx + 1 and k[1] >= col_start and k[1] < col_end:
                    cell.set_facecolor(colors[row_idx][col_idx])
                    col_idx += 1

    if special_rows_colors is not None:
        for special_rows_color in special_rows_colors:
            set_cells_facecolor(mpl_table,
                                row=special_rows_color[0],
                                color=special_rows_color[1],
                                alpha=0.5*alpha,
                                bold_font=bold_font)

    if special_columns_colors is not None:
        for special_columns_color in special_columns_colors:
            set_cells_facecolor(mpl_table,
                                column=special_columns_color[0],
                                color=special_columns_color[1],
                                alpha=0.5*alpha,
                                bold_font=bold_font,
                                header_row_id=header_row_id)

    if diagonal_color is not None:
        set_diag_cells_facecolor(mpl_table, color=diagonal_color)

    if data_colors is not None:
        set_data_colors(mpl_table, header_column_id=0 if add_index_as_column else -1,
                        data_colors=data_colors)

    set_row_edge_color(mpl_table, row=None, color='lightgray')

    if left_aligned_first_col:
        set_align_for_column(mpl_table, col=0, align='left')
    else:
        set_align_for_column(mpl_table, col=0, align='right')

    if rows_edge_lines is not None or columns_edge_lines is not None:
        ax.axis(xmin=0, xmax=df.shape[1], ymin=df.shape[0], ymax=-1)  # need to reset axis to match rows position

    if rows_edge_lines is not None:
        for rows_edge_line in rows_edge_lines:
            ax.axhline(y=rows_edge_line, color='black', alpha=0.5*alpha)

    if columns_edge_lines is not None:
        for columns_edge_line in columns_edge_lines:
            ax.axvline(x=columns_edge_line[0], color=columns_edge_line[1], alpha=0.5*alpha)
            # set_column_edge_color(mpl_table, column=columns_edge_line[0], color=columns_edge_line[1])

    if title is not None:
        put.set_title(ax=ax, title=title, fontsize=fontsize, **kwargs)

    return fig


def plot_df_table_with_ci(df: pd.DataFrame,
                          df_ci: pd.DataFrame,
                          var_format: str = '{:.2f}',
                          ax: plt.Subplot = None,
                          is_add_heatmap: bool = True,
                          axis: Literal[None, 0, 1] = 1,  # heatmap by column
                          **kwargs
                          ) -> plt.Figure:
    """
    table with ci
    """
    table_str = dfs.df_with_ci_to_str(df=df,
                                      df_ci=df_ci,
                                      var_format=var_format)
    if is_add_heatmap:
        data_colors = put.compute_heatmap_colors(a=df.to_numpy(), axis=axis)
    else:
        data_colors = None

    fig = plot_df_table(df=table_str,
                        data_colors=data_colors,
                        ax=ax,
                        **kwargs)
    return fig


def set_row_edge_color(table: Table,
                       row: int = None,
                       color: str = 'slategray'
                       ) -> None:
    for k, cell in table._cells.items():
        if row is None:
            cell.set_edgecolor(color)
        else:
            if k[0] == row:
                cell.set_edgecolor(color)


def set_column_edge_color(table: Table,
                          column: int = None,
                          color: str = 'slategray'
                          ) -> None:
    for k, cell in table._cells.items():
        if column is None:
            cell.set_edgecolor(color)
        else:
            if k[1] == column:
                cell.set_edgecolor(color)
                cell.visible_edges = "L"


def set_cells_facecolor(table: Table,
                        header_row_id: int = 0,
                        row: int = None,
                        column: int = None,
                        color: str = 'slategray',
                        alpha: float = 1.0,
                        bold_font: bool = False
                        ) -> None:

    for k, cell in table._cells.items():
        if row is None and column is None:
            cell.set_facecolor(color)
            if bold_font:
                txt = cell.get_text()
                txt.set_fontweight("bold")

        elif column is None:  # set solor to row
            if k[0] == row:
                cell.set_facecolor(color)
                if bold_font:
                    txt = cell.get_text()
                    txt.set_fontweight("bold")

        elif row is None:  # set solor to column
            if k[1] == column and k[0] > header_row_id:
                cell.set_facecolor(color)
                cell.set_alpha(alpha)
                if bold_font:
                    txt = cell.get_text()
                    txt.set_fontweight("bold")


def set_diag_cells_facecolor(table: Table,
                             color: str = 'slategray',
                             bold_font: bool = False
                             ) -> None:
    for k, cell in table._cells.items():
        if k[0] == k[1] and k[0] > 0:
            cell.set_facecolor(color)
            if bold_font:
                txt = cell.get_text()
                txt.set_fontweight("bold")


def set_data_colors(table: Table,
                    data_colors: np.ndarray,
                    header_row_id: int = 0,
                    header_column_id = 0,
                    bold_font: bool = False
                    ) -> None:
    for k, cell in table._cells.items():
        if k[1] > header_column_id and k[0] > header_row_id:
            cell.set_facecolor(data_colors[k[0]-1][k[1]-1])
            if bold_font:
                txt = cell.get_text()
                txt.set_fontweight("bold")


def set_align_for_column(table: Table,
                         col: int,
                         align: str = 'left'
                         ) -> None:
    cells = [key for key in table._cells if key[1] == col]
    for cell in cells:
        table.properties()["celld"][cell]._loc = align


class UnitTests(Enum):
    TABLE = 0


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.TABLE:

        cars = {'Brand': ['Honda Civic', 'Toyota Corolla', 'Ford Focus', 'Audi A4'],
                'Price': [220.0, 250.0, 270.0, 35.0],
                'Engine': [175.0, 300.0, 100.0, 500.0],
                'Speed': [200.0, 150.0, 200.0, 175.0]}

        data = pd.DataFrame.from_dict(cars)
        data = data.set_index('Brand', drop=False)
        print(data)
        data['Price'] = data['Price']
        data['Engine'] = data['Engine']
        plot_df_table(df=data, heatmap_columns=[2], bold_font=False)
        plot_df_table(df=data, heatmap_rows_columns=((0, len(data.index)), (3, 4)))

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.TABLE

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
Optimizers"""

from __future__ import division
import numpy as np
from scipy.optimize import minimize
from typing import List
from enum import Enum

import qis.utils.dates as da


def calculate_portfolio_var(w: np.ndarray, covar: np.ndarray) -> float:
    return w @ covar @ w.T


def calculate_risk_contribution(w: np.ndarray, covar: np.ndarray) -> np.ndarray:
    portfolio_vol = np.sqrt(calculate_portfolio_var(w, covar))
    marginal_risk_contribution = covar @ w.T
    rc = np.multiply(marginal_risk_contribution, w) / portfolio_vol
    return rc


def calculate_diversification_ratio(w: np.ndarray, covar: np.ndarray) -> float:
    avg_weighted_vol = np.sqrt(np.diag(covar)) @ w.T
    portfolio_vol = np.sqrt(calculate_portfolio_var(w, covar))
    diversification_ratio = avg_weighted_vol/portfolio_vol
    return diversification_ratio


def max_diversification_objective(w: np.ndarray, pars: List[np.ndarray]) -> float:
    covar = pars[0]
    return -calculate_diversification_ratio(w=w, covar=covar)


def carra_objective(w: np.ndarray, pars: List[np.ndarray]) -> float:
    means, covar, carra = pars[0], pars[1], pars[2]
    v = means.T @ w - 0.5*carra*w.T @ covar @ w
    return -v


def carra_objective_exp(w: np.ndarray, pars: List[np.ndarray]) -> float:
    means, covar, carra = pars[0], pars[1], pars[2]
    v = np.exp(-carra*means.T @ w + 0.5*carra*carra*w.T @ covar @ w)
    return v


def carra_objective_mixture(w: np.ndarray, pars: List[np.ndarray]) -> float:
    means, covars, probs, carra = pars[0], pars[1], pars[2], pars[3]
    v = 0.0
    for idx, prob in enumerate(probs):
        v = v + prob*np.exp(-carra*means[idx].T @ w + 0.5*carra*carra*w.T @ covars[idx] @ w)
    return v


def risk_budget_objective(x, pars):
    covar, budget = pars[0], pars[1]
    asset_rc = calculate_risk_contribution(x, covar)
    sig_p = np.sqrt(calculate_portfolio_var(x, covar))
    if budget is not None:
        risk_target = np.where(np.isnan(budget), asset_rc, np.multiply(sig_p, budget))  # budget can be nan f
    else:
        risk_target = np.ones_like(asset_rc) / asset_rc.shape[0]
    sse = np.nansum(np.square(asset_rc - risk_target))
    return sse


def risk_budget_objective_mod(x, pars):
    covar, budget = pars[0], pars[1]
    sig_p = np.sqrt(calculate_portfolio_var(x, covar))  # portfolio sigma
    risk_target = np.asmatrix(np.multiply(sig_p, budget))
    asset_RC = calculate_risk_contribution(x, covar)
    sse = np.sum(np.square(asset_RC[:-1] - risk_target.T[:-1] - np.mean(asset_RC[:-1] - risk_target.T[:-1])))
    return sse


def total_weight_constraint(x, total: float = 1.0):
    return total - np.sum(x)


def sum_of_log_weight_constraint(x, risk_budget):
    return np.log(x).dot(risk_budget)


def long_only_constraint(x):
    return x


def portfolio_volatility_min(x, covar, target_vol, freq_vol, af: float = 12.0):
    vol_dt = np.sqrt(af)
    return vol_dt * np.sqrt(calculate_portfolio_var(x, covar)) - (target_vol - 0.001)


def portfolio_volatility_max(x, V, target_vol, freq_vol, af: float = 12.0):
    vol_dt = np.sqrt(af)
    return - (vol_dt * np.sqrt(calculate_portfolio_var(x, V)) - (target_vol + 0.001))


def solve_equal_risk_contribution(covar: np.ndarray,
                                  budget: np.ndarray = None,
                                  weight_mins: np.ndarray = None,
                                  weight_maxs: np.ndarray = None,
                                  is_gross_notional_one: bool = True,
                                  disp: bool = False
                                  ) -> np.ndarray:
    n = covar.shape[0]
    x0 = np.ones(n) / n

    cons = [{'type': 'ineq', 'fun': long_only_constraint}]
    if is_gross_notional_one:
        cons.append({'type': 'eq', 'fun': total_weight_constraint})
    if weight_mins is not None:
        cons.append({'type': 'ineq', 'fun': lambda x: x - weight_mins})
    if weight_maxs is not None:
        cons.append({'type': 'ineq', 'fun': lambda x: weight_maxs - x})

    res = minimize(risk_budget_objective, x0, args=[covar, budget], method='SLSQP', constraints=cons,
                   options={'disp': disp, 'ftol': 1e-18, 'maxiter': 200})
    w_rb = res.x

    print(f'sigma_p = {np.sqrt(calculate_portfolio_var(w_rb, covar))}, weights: {w_rb}, '
          f'risk contrib.s: {calculate_risk_contribution(w_rb, covar).T} '
          f'sum of weights: {sum(w_rb)}')
    return w_rb


def solve_max_diversification(covar: np.ndarray,
                              weight_mins: np.ndarray = None,
                              weight_maxs: np.ndarray = None,
                              is_gross_notional_one: bool = True,
                              disp: bool = False,
                              ) -> np.ndarray:
    n = covar.shape[0]
    x0 = np.ones(n) / n

    cons = [{'type': 'ineq', 'fun': long_only_constraint}]
    if is_gross_notional_one:
        cons.append({'type': 'eq', 'fun': total_weight_constraint})
    if weight_mins is not None:
        cons.append({'type': 'ineq', 'fun': lambda x: x - weight_mins})
    if weight_maxs is not None:
        cons.append({'type': 'ineq', 'fun': lambda x: weight_maxs - x})

    res = minimize(max_diversification_objective, x0, args=[covar], method='SLSQP', constraints=cons,
                   options={'disp': disp, 'ftol': 1e-18, 'maxiter': 200})
    w_rb = res.x

    print(f'sigma_p = {np.sqrt(calculate_portfolio_var(w_rb, covar))}, weights: {w_rb}, '
          f'risk contrib.s: {calculate_risk_contribution(w_rb, covar).T} '
          f'sum of weights: {sum(w_rb)}')
    return w_rb


def solve_risk_parity_alt(covar: np.ndarray,
                          budget: np.ndarray = None,
                          disp: bool = False):
    n = covar.shape[0]
    if budget is None:
        budget = np.ones(n) / n
    x0 = budget
    cons = [{'type': 'ineq', 'fun': long_only_constraint},
            {'type': 'eq', 'fun': total_weight_constraint},
            {'type': 'ineq', 'fun': sum_of_log_weight_constraint, 'args': (budget,)}]

    res = minimize(calculate_portfolio_var, x0, args=covar, method='SLSQP', constraints=cons,
                   options={'disp': disp, 'ftol': 1e-14})

    w_rb = res.x

    print(f'(ALT) sigma_p = {np.sqrt(calculate_portfolio_var(w_rb, covar))}, weights: {w_rb}, '
          f'risk contrib.s: {calculate_risk_contribution(w_rb, covar).T} '
          f'sum of weights: {sum(w_rb)}')
    return w_rb


def solve_risk_parity_constr_vol(covar: np.ndarray,
                                 target_vol: float = None,
                                 disp: bool = False):
    n = covar.shape[0]
    budget = np.ones(n) / n
    x0 = budget

    cons = [{'type': 'ineq', 'fun': long_only_constraint},
            {'type': 'eq', 'fun': total_weight_constraint},
            {'type': 'ineq', 'fun': portfolio_volatility_min, 'args': (covar, target_vol)},
            {'type': 'ineq', 'fun': portfolio_volatility_max, 'args': (covar, target_vol)}]

    res = minimize(risk_budget_objective_mod, x0, args=[covar, budget], method='SLSQP', constraints=cons,
                   options={'disp': disp, 'ftol': 1e-14})

    w_rb = res.x

    print(f'(CON) sigma_p = {np.sqrt(calculate_portfolio_var(w_rb, covar))}, weights: {w_rb}, '
          f'risk contrib.s: {calculate_risk_contribution(w_rb, covar).T} '
          f'sum of weights: {sum(w_rb)}')
    return w_rb


def solve_cara(means: np.ndarray,
               covar: np.ndarray,
               carra: float = 0.5,
               weight_mins: np.ndarray = None,
               weight_maxs: np.ndarray = None,
               disp: bool = False,
               is_exp: bool = False
               ) -> np.ndarray:
    n = covar.shape[0]
    x0 = np.ones(n) / n
    cons = [{'type': 'ineq', 'fun': long_only_constraint},
            {'type': 'eq', 'fun': total_weight_constraint}]
    if weight_mins is not None:
        cons.append({'type': 'ineq', 'fun': lambda x: x - weight_mins})
    if weight_maxs is not None:
        cons.append({'type': 'ineq', 'fun': lambda x: weight_maxs - x})

    if is_exp:
        func = carra_objective_exp
    else:
        func = carra_objective
    res = minimize(func, x0, args=[means, covar, carra], method='SLSQP', constraints=cons,
                   options={'disp': disp, 'ftol': 1e-16})
    w_rb = res.x

    print(f'return_p = {w_rb@means}, '
          f'sigma_p = {np.sqrt(calculate_portfolio_var(w_rb, covar))}, weights: {w_rb}, '
          f'risk contrib.s: {calculate_risk_contribution(w_rb, covar).T} '
          f'sum of weights: {sum(w_rb)}')
    return w_rb


def solve_cara_mixture(means: List[np.ndarray],
                       covars: List[np.ndarray],
                       probs: np.ndarray,
                       carra: float = 0.5,
                       weight_mins: np.ndarray = None,
                       weight_maxs: np.ndarray = None,
                       disp: bool = False,
                       is_print_log: bool = True
                       ) -> np.ndarray:
    n = means[0].shape[0]
    x0 = np.ones(n) / n
    cons = [{'type': 'ineq', 'fun': long_only_constraint},
            {'type': 'eq', 'fun': total_weight_constraint}]
    if weight_mins is not None:
        cons.append({'type': 'ineq', 'fun': lambda x: x - weight_mins})
    if weight_maxs is not None:
        cons.append({'type': 'ineq', 'fun': lambda x: weight_maxs - x})

    res = minimize(carra_objective_mixture, x0, args=[means, covars, probs, carra], method='SLSQP',
                   constraints=cons,
                   options={'disp': disp, 'ftol': 1e-16})
    w_rb = res.x

    if is_print_log:
        weights_str = ' '.join([f"{w:0.3f}" for w in w_rb])
        print(weights_str)
    return w_rb


class UnitTests(Enum):
    RISK_PARITY = 1
    CARA = 2
    CARA_MIX = 3


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.RISK_PARITY:
        covar = np.array([[0.2 ** 2, 0.015, 0.0],
                          [0.015, 0.15 ** 2, 0.0],
                          [0.0, 0.0, 0.1]])
        w_rb = solve_equal_risk_contribution(covar=covar, disp=True)
        w_md = solve_max_diversification(covar=covar, disp=True)

        w_rb = solve_equal_risk_contribution(covar=covar, is_gross_notional_one=False, disp=True)
        w_md = solve_max_diversification(covar=covar, is_gross_notional_one=False, disp=True)

    elif unit_test == UnitTests.CARA:
        means = np.array([0.3, 0.1])
        covar = np.array([[0.2 ** 2, 0.01],
                          [0.01, 0.1 ** 2]])
        w_rb = solve_cara(means=means, covar=covar, carra=10, is_exp=False, disp=True)
        w_rb = solve_cara(means=means, covar=covar, carra=10, is_exp=True, disp=True)

    elif unit_test == UnitTests.CARA_MIX:
        means = [np.array([0.05, -0.1]), np.array([0.05, 2.0])]
        covars = [np.array([[0.2 ** 2, 0.01],
                          [0.01, 0.2 ** 2]]),
                 np.array([[0.2 ** 2, 0.01],
                           [0.01, 0.2 ** 2]])
                 ]
        probs = np.array([0.95, 0.05])
        w_rb = solve_cara_mixture(means=means, covars=covars, probs=probs, carra=20.0, disp=True)


if __name__ == '__main__':

    unit_test = UnitTests.RISK_PARITY

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



# packages
import matplotlib.pyplot as plt
import seaborn as sns
import yfinance as yf
from enum import Enum

# qis
import qis


class UnitTests(Enum):
    PLOT_MIXURE = 1


def run_unit_test(unit_test: UnitTests):

    perf_params = qis.PerfParams(freq='W-WED')
    kwargs = dict(fontsize=12, digits_to_show=1, sharpe_digits=2,
                  alpha_format='{0:+0.0%}',
                  beta_format='{:0.1f}',
                  alpha_an_factor=12)

    time_periods = [qis.TimePeriod('31Aug2002', '31Dec2019'), qis.TimePeriod('31Dec2019', '16Dec2022')]

    if unit_test == UnitTests.PLOT_MIXURE:

        prices = yf.download(tickers=['SPY', 'TLT'], start=None, end=None)['Adj Close'].dropna()

        n_components = 3

        with sns.axes_style('white'):
            fig1, axs = plt.subplots(1, len(time_periods), figsize=(15, 5), constrained_layout=True)

        for idx, time_period in enumerate(time_periods):
            prices_ = time_period.locate(prices)
            rets = qis.to_returns(prices=prices_, is_log_returns=True, drop_first=True, freq=perf_params.freq)
            params = qis.fit_gaussian_mixture(x=rets.to_numpy(), n_components=n_components, idx=1)
            qis.plot_mixure2(x=rets.to_numpy(),
                            n_components=n_components,
                            columns=prices.columns,
                            title=f"({idx+1}) Returns and ellipsoids of Gaussian clusters for period {time_period.to_str()}",
                            ax=axs[idx],
                            **kwargs)

            means, vols, corrs = params.get_all_params(columns=prices.columns, vol_scaler=12.0)
            print(f"means=\n{means}")
            print(f"vols=\n{vols}")
            print(f"corrs=\n{corrs}")

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.PLOT_MIXURE

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
multi asset report is generated for a set of asset prices
with comparison to 1-2 benchmarks
output is one-page figure with key numbers
"""
# packages
import pandas as pd
import matplotlib.pyplot as plt
from typing import Union, List, Optional, Tuple
from enum import Enum

# qis
import qis
from qis import TimePeriod
import qis.utils as qu
import qis.perfstats as qs
import qis.plots as qp
import qis.models as qm
from qis.perfstats import PerfStat, PerfParams, BenchmarkReturnsQuantileRegimeSpecs, RegimeData


PERF_COLUMNS = (
    # PerfStat.START_DATE,
    # PerfStat.END_DATE,
    PerfStat.TOTAL_RETURN,
    PerfStat.PA_RETURN,
    PerfStat.VOL,
    PerfStat.SHARPE,
    PerfStat.BEAR_SHARPE,
    PerfStat.NORMAL_SHARPE,
    PerfStat.BULL_SHARPE,
    PerfStat.MAX_DD,
    PerfStat.MAX_DD_VOL,
    PerfStat.WORST,
    PerfStat.BEST,
    PerfStat.SKEWNESS,
    PerfStat.KURTOSIS)


# default perf and regime params
PERF_PARAMS = PerfParams(freq='W-WED')
REGIME_PARAMS = BenchmarkReturnsQuantileRegimeSpecs(freq='Q')


class MultiAssetReport:

    def __init__(self,
                 prices: pd.DataFrame,
                 benchmark_prices: Union[pd.Series, pd.DataFrame],
                 perf_params: PerfParams = PERF_PARAMS,
                 regime_params: qs.BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS
                 ):

        # make sure it is consistent
        self.prices = prices
        self.benchmark_prices = benchmark_prices.reindex(index=prices.index, method='ffill')
        self.perf_params = perf_params
        self.regime_params = regime_params

    def get_prices(self, benchmark: str = None, time_period: TimePeriod = None) -> pd.DataFrame:
        if benchmark is not None and benchmark not in self.prices.columns:
            if isinstance(self.benchmark_prices, pd.Series):
                prices = pd.concat([self.benchmark_prices, self.prices], axis=1)
            else:
                prices = pd.concat([self.benchmark_prices[benchmark], self.prices], axis=1)
        else:
            prices = self.prices
        if time_period is not None:
            prices = time_period.locate(prices)
        return prices

    def add_regime_shadows(self, ax: plt.Subplot, regime_benchmark_str: str, data_df: pd.DataFrame) -> None:
        if isinstance(self.benchmark_prices, pd.Series):
            pivot_prices = self.benchmark_prices
        else:
            if regime_benchmark_str is None:
                regime_benchmark_str = self.benchmark_prices.columns[0]
            pivot_prices = self.benchmark_prices[regime_benchmark_str]
        pivot_prices = pivot_prices.reindex(index=data_df.index, method='ffill')
        qp.add_bnb_regime_shadows(ax=ax,
                                  data_df=data_df,
                                  pivot_prices=pivot_prices,
                                  benchmark=regime_benchmark_str,
                                  regime_params=self.regime_params)

    def plot_ra_perf_table(self,
                           benchmark: str,
                           time_period: TimePeriod = None,
                           ax: plt.Subplot = None,
                           **kwargs) -> None:
        prices = self.get_prices(benchmark, time_period=time_period)
        qp.plot_ra_perf_table_benchmark(prices=prices,
                                        benchmark=benchmark,
                                        perf_params=self.perf_params,
                                        perf_columns=qs.BENCHMARK_TABLE_COLUMNS,
                                        title=f"RA performance table: {qu.get_time_period(prices).to_str()}",
                                        rotation_for_columns_headers=0,
                                        ax=ax,
                                        **kwargs)

    def plot_ra_regime_table(self,
                             regime_benchmark_str: str = None,
                             time_period: TimePeriod = None,
                             perf_columns: List[PerfStat] = PERF_COLUMNS,
                             columns_title: str = 'Programs',
                             first_column_width: float = 3.5,
                             ax: plt.Subplot = None,
                             **kwargs) -> None:

        prices = pd.concat([self.prices, self.benchmark_prices], axis=1)

        if time_period is not None:
            prices = time_period.locate(prices)

        cvar_table = qs.compute_bnb_regimes_pa_perf_table(prices=prices,
                                                          benchmark=regime_benchmark_str,
                                                          perf_params=self.perf_params,
                                                          regime_params=self.regime_params)
        table_data = pd.DataFrame(data=prices.columns, index=cvar_table.index, columns=[columns_title])

        for perf_column in perf_columns:
            table_data[perf_column.to_str()] = qu.series_to_str(ds=cvar_table[perf_column.to_str()],
                                                             var_format=perf_column.to_format(**kwargs))

        special_columns_colors = [(0, 'steelblue')]
        qp.plot_df_table(df=table_data,
                         first_column_width=first_column_width,
                         add_index_as_column=False,
                         index_column_name='Strategies',
                         special_columns_colors=special_columns_colors,
                         ax=ax,
                         **kwargs)

    def plot_nav(self,
                 regime_benchmark_str: str = None,
                 var_format: str = '{:.0%}',
                 sharpe_format: str = '{:.2f}',
                 title: str = 'Cumulative performance',
                 is_log: bool = True,
                 time_period: TimePeriod = None,
                 ax: plt.Subplot = None,
                 **kwargs) -> None:
        prices = self.get_prices(time_period=time_period, benchmark=regime_benchmark_str)
        qp.plot_prices(prices=prices,
                       perf_params=self.perf_params,
                       start_to_one=True,
                       is_log=is_log,
                       var_format=var_format,
                       sharpe_format=sharpe_format,
                       title=title,
                       ax=ax,
                       **kwargs)
        self.add_regime_shadows(ax=ax, regime_benchmark_str=regime_benchmark_str, data_df=prices)

    def plot_drawdowns(self,
                       regime_benchmark_str: str = None,
                       time_period: TimePeriod = None,
                       ax: plt.Subplot = None,
                       **kwargs) -> None:
        prices = self.get_prices(time_period=time_period, benchmark=regime_benchmark_str)
        qp.plot_drawdown(prices=prices, ax=ax, **kwargs)
        self.add_regime_shadows(ax=ax, regime_benchmark_str=regime_benchmark_str, data_df=prices)

    def plot_rolling_time_under_water(self,
                                      regime_benchmark_str: str = None,
                                      time_period: TimePeriod = None,
                                      ax: plt.Subplot = None,
                                      **kwargs) -> None:
        prices = self.get_prices(time_period=time_period, benchmark=regime_benchmark_str)
        qp.plot_rolling_time_under_water(prices=prices, ax=ax, **kwargs)
        self.add_regime_shadows(ax=ax, regime_benchmark_str=regime_benchmark_str, data_df=prices)

    def plot_annual_returns(self,
                            heatmap_freq: str = 'A',
                            date_format: str = '%Y',
                            time_period: TimePeriod = None,
                            ax: plt.Subplot = None,
                            **kwargs) -> None:
        local_kwargs = qu.update_kwargs(kwargs=kwargs,
                                        new_kwargs=dict(fontsize=4,
                                                        square=False,
                                                        x_rotation=90))
        qp.plot_periodic_returns_table(prices=self.get_prices(time_period=time_period),
                                       freq=heatmap_freq,
                                       ax=ax,
                                       title=f"Annual Returns",
                                       date_format=date_format,
                                       **local_kwargs)

    def plot_corr_table(self,
                        freq: str = 'W-WED',
                        time_period: TimePeriod = None,
                        ax: plt.Subplot = None,
                        **kwargs) -> None:
        local_kwargs = qu.update_kwargs(kwargs=kwargs, new_kwargs=dict(fontsize=4))
        prices = self.get_prices(time_period=time_period)
        qm.plot_corr_table(prices=prices,
                           x_rotation=90,
                           freq=freq,
                           title=f"Correlation {freq} returns: {qu.get_time_period(prices).to_str()}",
                           ax=ax,
                           **local_kwargs)

    def plot_returns_scatter(self,
                             benchmark: str,
                             time_period: TimePeriod = None,
                             ax: plt.Subplot = None,
                             **kwargs) -> None:
        local_kwargs = qu.update_kwargs(kwargs=kwargs,
                                        new_kwargs=dict(weight='bold',
                                                        markersize=8,
                                                        x_rotation=0,
                                                        first_color_fixed=False,
                                                        ci=None))
        qp.plot_returns_scatter(prices=self.get_prices(benchmark=benchmark, time_period=time_period),
                                benchmark=benchmark,
                                freq=self.perf_params.freq_reg,
                                order=2,
                                title=f"Scatterplot of {self.perf_params.freq_reg}-returns vs {benchmark}",
                                ax=ax,
                                **local_kwargs)

    def plot_benchmark_beta(self,
                            benchmark: str,
                            freq: str = 'B',
                            span: int = 60,
                            time_period: TimePeriod = None,
                            ax: plt.Subplot = None,
                            **kwargs) -> None:
        returns = qs.to_returns(prices=self.get_prices(benchmark=benchmark), freq=freq)
        ewm_linear_model = qm.estimate_ewm_linear_model(x=returns[benchmark].to_frame(),
                                                        y=returns.drop(benchmark, axis=1),
                                                        span=span,
                                                        is_x_correlated=True)
        ewm_linear_model.plot_factor_loadings(factor=benchmark,
                                              time_period=time_period,
                                              title=f"Rolling EWM span-{span:0.0f} beta to {benchmark}",
                                              ax=ax,
                                              **kwargs)
        self.add_regime_shadows(ax=ax, regime_benchmark_str=benchmark, data_df=self.prices)

    def plot_regime_data(self,
                         benchmark: str,
                         regime_data_to_plot: RegimeData = RegimeData.REGIME_SHARPE,
                         time_period: TimePeriod = None,
                         var_format: Optional[str] = None,
                         is_conditional_sharpe: bool = True,
                         legend_loc: Optional[str] = 'upper center',
                         ax: plt.Subplot = None,
                         **kwargs) -> None:
        prices = self.get_prices(time_period=time_period)
        title = f"Sharpe ratio decomposition by Strategies to {benchmark} Bear/Normal/Bull regimes"
        regime_classifier = qs.BenchmarkReturnsQuantilesRegime(regime_params=self.regime_params)
        fig = qp.plot_regime_data(regime_classifier=regime_classifier,
                                  prices=prices,
                                  benchmark=benchmark,
                                  is_conditional_sharpe=is_conditional_sharpe,
                                  regime_data_to_plot=regime_data_to_plot,
                                  var_format=var_format or '{:.2f}',
                                  legend_loc=legend_loc,
                                  perf_params=self.perf_params,
                                  title=title,
                                  ax=ax,
                                  **kwargs)


def generate_multi_asset_factsheet(prices: pd.DataFrame,
                                   benchmark_prices: Union[pd.Series, pd.DataFrame] = None,
                                   benchmark: str = None,
                                   perf_params: PerfParams = PERF_PARAMS,
                                   regime_params: qs.BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                                   heatmap_freq: str = 'A',
                                   time_period: TimePeriod = None,  # time period for reporting
                                   figsize: Tuple[float, float] = (8.3, 11.7),  # A4 for portrait
                                   **kwargs
                                   ) -> plt.Figure:

    # add default benchmark str
    if benchmark is None and benchmark_prices is not None:
        if benchmark_prices is None:
            raise ValueError(f"pass either benchmark or benchmark_prices")
        else:
            if isinstance(benchmark_prices, pd.Series):
                benchmark = benchmark_prices.name
            else:
                benchmark = benchmark_prices.columns[0]
    if benchmark is not None and benchmark_prices is None:
        if benchmark in prices.columns:
            benchmark_prices = prices[benchmark]
        else:
            raise ValueError(f"benchmark must be in prices")

    # report data
    report = MultiAssetReport(prices=prices,
                              benchmark_prices=benchmark_prices,
                              perf_params=perf_params,
                              regime_params=regime_params)

    local_kwargs = dict(fontsize=5,
                        linewidth=0.5,
                        weight='normal',
                        markersize=1,
                        framealpha=0.75,
                        x_date_freq='A',
                        time_period=time_period)
    # overrite local_kwargs with kwargs is they are provided
    kwargs = qu.update_kwargs(local_kwargs, kwargs)

    # figure
    fig = plt.figure(figsize=figsize, constrained_layout=True)
    gs = fig.add_gridspec(nrows=8, ncols=4, wspace=0.0, hspace=0.0)

    report.plot_nav(regime_benchmark_str=benchmark,
                    ax=fig.add_subplot(gs[:2, :2]),
                    **kwargs)

    report.plot_drawdowns(regime_benchmark_str=benchmark,
                          ax=fig.add_subplot(gs[2:4, :2]),
                          **kwargs)

    report.plot_rolling_time_under_water(regime_benchmark_str=benchmark,
                                         ax=fig.add_subplot(gs[4:6, :2]),
                                         **kwargs)

    report.plot_benchmark_beta(benchmark=benchmark,
                               ax=fig.add_subplot(gs[6:8, :2]),
                               **kwargs)

    report.plot_ra_perf_table(benchmark=benchmark,
                              ax=fig.add_subplot(gs[0, 2:]),
                              **kwargs)

    time_period1 = qu.get_time_period_shifted_by_years(time_period=qu.get_time_period(df=prices))
    report.plot_ra_perf_table(benchmark=benchmark,
                              ax=fig.add_subplot(gs[1, 2:]),
                              **qu.update_kwargs(kwargs, dict(time_period=time_period1)))

    report.plot_annual_returns(ax=fig.add_subplot(gs[2:4, 2:]),
                               heatmap_freq=heatmap_freq,
                               **kwargs)

    report.plot_corr_table(freq=report.perf_params.freq,
                           ax=fig.add_subplot(gs[4, 2]),
                           **kwargs)
    report.plot_corr_table(freq=report.perf_params.freq,
                           ax=fig.add_subplot(gs[4, 3]),
                           **qu.update_kwargs(kwargs, dict(time_period=time_period1)))

    report.plot_regime_data(benchmark=benchmark,
                            ax=fig.add_subplot(gs[5, 2:]),
                            **kwargs)

    report.plot_returns_scatter(benchmark=benchmark,
                                ax=fig.add_subplot(gs[6:8, 2:]),
                                **kwargs)
    return fig


class UnitTests(Enum):
    FACTSHEET = 1


def run_unit_test(unit_test: UnitTests):

    from qis.test_data import load_etf_data
    prices = load_etf_data().dropna()

    if unit_test == UnitTests.FACTSHEET:
        fig = generate_multi_asset_factsheet(prices=prices,
                                             benchmark='SPY',
                                             heatmap_freq='A',
                                             perf_params=PERF_PARAMS,
                                             regime_params=REGIME_PARAMS)
        qis.save_figs_to_pdf(figs=[fig],
                             file_name=f"multiasset_report", orientation='landscape',
                             local_path=qis.local_path.get_output_path())

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.FACTSHEET

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



# packages
import pandas as pd
import matplotlib.pyplot as plt
from typing import Union, Tuple, Optional
from enum import Enum

# qis
import qis.plots.utils as put
import qis.plots.time_series as pts
import qis.perfstats.perf_stats as pt


class DdLegendType(Enum):
    NONE = 1
    SIMPLE = 2
    DETAILED = 3


def plot_drawdown(prices: pd.DataFrame,
                  title: Optional[str] = None,
                  var_format: str = '{:.0%}',
                  dd_legend_type: DdLegendType = DdLegendType.DETAILED,
                  legend_loc: str = 'lower left',
                  y_limits: Tuple[Optional[float], Optional[float]] = (None, 0.0),
                  ax: plt.Subplot = None,
                  **kwargs
                  ) -> plt.Figure:

    max_dd_data = pt.compute_drawdown_data(prices=prices)

    if dd_legend_type == DdLegendType.NONE:
        legend_loc = None
        legend_labels = None
    else:
        legend_labels = []
        for column in max_dd_data.columns:
            avg, quant, nmax, last = pt.compute_avg_max_dd(ds=max_dd_data[column], is_max=False)
            if dd_legend_type == DdLegendType.SIMPLE:
                legend_labels.append(f"{column}, max dd={var_format.format(nmax)}")
            elif dd_legend_type == DdLegendType.DETAILED:
                legend_labels.append(f"{column}, mean={var_format.format(avg)},"
                                     f" quantile_10%={var_format.format(quant)}, max={var_format.format(nmax)},"
                                     f" last={var_format.format(last)}")
            else:
                raise NotImplementedError(f"{dd_legend_type}")

    fig = pts.plot_time_series(df=max_dd_data,
                               var_format=var_format,
                               legend_loc=legend_loc,
                               legend_labels=legend_labels,
                               title=title,
                               y_limits=y_limits,
                               ax=ax,
                               **kwargs)
    return fig


def plot_rolling_time_under_water(prices: pd.DataFrame,
                                  title: Union[str, None] = None,
                                  var_format: str = '{:,.0f}',
                                  y_limits: Tuple[Optional[float], Optional[float]] = (0.0, None),
                                  ax: plt.Subplot = None,
                                  **kwargs
                                  ) -> plt.Figure:
    if isinstance(prices, pd.Series):
        prices = prices.to_frame()

    max_dd_data, time_under_water = pt.compute_drawdown_time_under_water(prices=prices)

    legend_labels = []
    for column in time_under_water.columns:
        avg, quant, nmax, last = pt.compute_avg_max_dd(ds=time_under_water[column], is_max=True)
        legend_labels.append(f"{column}, mean={var_format.format(avg)}, "
                             f"quantile_10%={var_format.format(quant)}, max={var_format.format(nmax)},"
                             f" last={var_format.format(last)}")

    fig = pts.plot_time_series(df=time_under_water,
                               var_format=var_format,
                               legend_loc='upper left',
                               legend_labels=legend_labels,
                               title=title,
                               y_limits=y_limits,
                               ax=ax,
                               **kwargs)
    return fig


def plot_top_drawdowns_paths(price: pd.Series,
                             freq: Optional[str] = 'D',
                             max_num: int = 10,
                             date_format: str = '%d%b%Y',
                             title: Union[str, None] = None,
                             var_format: str = '{:.0%}',
                             legend_loc: str = 'lower left',
                             highlight_ongoing: bool = False,
                             x_limits: Tuple[Optional[float], Optional[float]] = (0.0, None),
                             y_limits: Tuple[Optional[float], Optional[float]] = (None, 0.0),
                             ax: plt.Subplot = None,
                             **kwargs
                             ) -> plt.Figure:

    if freq is not None:
        price = price.asfreq(freq, method='ffill')  # it will have nans
    df = pt.compute_drawdowns_stats_table(price=price, max_num=max_num)
    price_slices = {}
    points = {}
    for start, trough, end, max_dd, peak, days_dd in zip(df['start'], df['trough'], df['end'], df['max_dd'], df['peak'],
                                                         df['days_dd']):
        name = f"{start:{date_format}}-{end:{date_format}}: max_dd={max_dd:0.0%}, days_dd={days_dd:0.0f}"
        price_slices[name] = (price.loc[start:end] / peak - 1.0).reset_index(drop=True)
        points[name] = {trough: max_dd}
    price_slices = pd.DataFrame.from_dict(price_slices, orient='columns')
    # price_slices = price_slices.dropna(axis=0, how='all')  # drop rows with all nans = business days

    n = len(price_slices.columns)
    colors, linestyles = put.get_n_colors(n=n), None
    if highlight_ongoing:
        linestyles = ['dotted'] * n
        last_time = price.index[-2]  # shouldbe one tick back
        for idx, end in enumerate(df['end']):
            if end == last_time:
                dd_slice = price_slices.columns[idx]
                name = f"{dd_slice}-ongoing"
                colors[idx] = 'black'
                linestyles[idx] = 'solid'
                price_slices = price_slices.rename({dd_slice: name}, axis=1)
                break

    fig = pts.plot_time_series(df=price_slices,
                               var_format=var_format,
                               legend_loc=legend_loc,
                               linestyles=linestyles,
                               legend_stats=pts.LegendStats.NONE,
                               x_limits=x_limits,
                               y_limits=y_limits,
                               xlabel='Days in drawdown',
                               ylabel='% performance from the last peak',
                               title=title,
                               colors=colors,
                               ax=ax,
                               **kwargs)
    return fig

class UnitTests(Enum):
    DRAWDOWN_TS = 1
    ROLLING_TIME = 2
    PLOT_TOP_DRAWDOWNS = 3


def run_unit_test(unit_test: UnitTests):

    from qis.test_data import load_etf_data
    prices = load_etf_data().dropna()

    if unit_test == UnitTests.DRAWDOWN_TS:
        plot_drawdown(prices=prices)

    elif unit_test == UnitTests.ROLLING_TIME:
        plot_rolling_time_under_water(prices=prices)

    elif unit_test == UnitTests.PLOT_TOP_DRAWDOWNS:
        # plot_top_drawdowns_ts(price=prices['TLT'], freq='D')
        plot_top_drawdowns_paths(price=prices['TLT'], highlight_ongoing=True, freq='D')

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.PLOT_TOP_DRAWDOWNS

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



# packages
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
import scipy.stats as ss
from sklearn.mixture import GaussianMixture
from scipy.stats import bernoulli
from dataclasses import dataclass
from enum import Enum
from matplotlib.patches import Ellipse
from typing import List, Tuple, Union, Dict

# qis
import qis.plots.utils as put
import qis.utils.np_ops as nop

RANDOM_STATE = 3


@dataclass
class Params:
    means: List[np.ndarray]
    covars: List[np.ndarray]
    probs: np.ndarray

    def print(self):
        print(f"probs=\n{self.probs}")
        print(f"mus=\n{self.means}")
        print(f"sigmas=\n{self.covars}")

    def get_params(self, idx: int = 0) -> pd.DataFrame:
        means = np.array([mean[idx] for mean in self.means])
        std = np.array([np.sqrt(covar[idx][idx]) for covar in self.covars])
        probs = pd.Series(self.probs, name='Prob')
        means = pd.Series(means, name='Mean')
        std = pd.Series(std, name='Std')
        return pd.concat([probs, means, std], axis=1)

    def get_all_params(self, columns: List[str], vol_scaler: float = 1.0
                       ) -> Tuple[pd.DataFrame, pd.DataFrame, Union[pd.Series, Dict[str, pd.DataFrame]]]:
        probs = pd.Series(self.probs, name='Prob')
        means = [probs]
        vols = []
        for idx, column in enumerate(columns):
            means.append(pd.Series([vol_scaler*mean[idx] for mean in self.means], name=column))
            vols.append(pd.Series([np.sqrt(vol_scaler)*np.sqrt(covar[idx][idx]) for covar in self.covars], name=column))
        means = pd.concat(means, axis=1)
        means.index.name = 'cluster'
        vols = pd.concat(vols, axis=1)
        vols.index.name = 'cluster'
        if len(columns) == 2:
            corrs = pd.Series([covar[0][1] / np.sqrt(covar[0][0]*covar[1][1]) for covar in self.covars])
        else:
            corrs = {}
            for idx, covar in enumerate(self.covars):
                corrs[f"{idx} cluster"] = pd.DataFrame(nop.covar_to_corr(covar), index=columns, columns=columns)

        return means, vols, corrs


def fit_gaussian_mixture(x: np.ndarray, n_components: int = 2, scaler: float = 1.0, idx: int = None):
    gmm = GaussianMixture(n_components=n_components,
                          covariance_type='full',
                          random_state=RANDOM_STATE)
    gmm.fit(x)
    if idx is not None:
        order = gmm.means_.argsort(axis=0)[:, idx]
        gmm.means_ = gmm.means_[order]
        gmm.covariances_ = gmm.covariances_[order]
        gmm.weights_ = gmm.weights_[order]
        gmm.precisions_ = gmm.precisions_[order]
        gmm.precisions_cholesky_ = gmm.precisions_cholesky_[order]

    return Params(means=[scaler*x for x in gmm.means_],  # convert to lists
                  covars=[scaler*x for x in gmm.covariances_],
                  probs=gmm.weights_)


def draw_ellipse(position, covariance,
                 ax: plt.Subplot,
                 color: str = 'gray',
                 **kwargs) -> None:
    """Draw an ellipse with a given position and covariance"""

    # Convert covariance to principal axes
    if covariance.shape == (2, 2):
        U, s, Vt = np.linalg.svd(covariance)
        angle = np.degrees(np.arctan2(U[1, 0], U[0, 0]))
        width, height = 2 * np.sqrt(s)
    else:
        angle = 0
        width, height = 2 * np.sqrt(covariance)

    # Draw the Ellipse
    for nsig in range(1, 4):
        ax.add_patch(Ellipse(position, nsig * width, nsig * height,
                             angle,
                             color=color,
                             **kwargs))


def plot_mixure1(x: np.ndarray, n_components: int = 2, label=True,
                 columns: List[str] = None, ax=None):
    ax = ax or plt.gca()

    gmm = GaussianMixture(n_components=n_components, covariance_type='full', random_state=RANDOM_STATE)
    gmm.fit(x)

    x_ = np.linspace(np.min(x), np.max(x), 100)

    # find useful parameters
    mean = gmm.fit(x).means_
    covs = gmm.fit(x).covariances_
    weights = gmm.fit(x).weights_

    # create necessary things to plot
    x_axis = np.linspace(1.25*np.min(x), 1.25*np.max(x), 100)
    y_axis0 = ss.norm.pdf(x_axis, float(mean[0][0]), np.sqrt(float(covs[0][0][0]))) * weights[0]  # 1st gaussian
    y_axis1 = ss.norm.pdf(x_axis, float(mean[1][0]), np.sqrt(float(covs[1][0][0]))) * weights[1]  # 2nd gaussian
    ax.hist(x, 10, density=True, color='lightblue')
    ax.plot(x_axis, y_axis0, lw=3, c='C0')
    ax.plot(x_axis, y_axis1, lw=3, c='C1')
    ax.plot(x_axis, y_axis0+y_axis1, lw=3, c='C2', ls='dashed')


def plot_mixure2(x: np.ndarray,
                 n_components: int = 2,
                 label: str = 'Cluster',
                 title: str = None,
                 columns: List[str] = None,
                 ax: plt.Subplot = None,
                 var_format: str = '{:.0%}',
                 idx: int = 1,
                 **kwargs
                 ) -> None:

    if ax is None:
        ax = plt.subplots(1, 1)

    gmm = GaussianMixture(n_components=n_components, covariance_type='full', random_state=RANDOM_STATE).fit(x)
    order = gmm.means_.argsort(axis=0)[:, idx]
    gmm.means_ = gmm.means_[order]
    gmm.covariances_ = gmm.covariances_[order]
    gmm.weights_ = gmm.weights_[order]
    gmm.precisions_ = gmm.precisions_[order]
    gmm.precisions_cholesky_ = gmm.precisions_cholesky_[order]

    labels = gmm.predict(x)

    if columns is None:
        columns = [f"X{n+1}" for n in range(x.shape[1])]
    data = pd.DataFrame(x, columns=columns)
    data[label] = labels

    if n_components == 3:
        colors = ['red', 'orange', 'green']
    else:
        colors = put.get_n_colors(n=n_components, last_color_fixed=False)
    sns.scatterplot(data=data,
                    x=columns[0],
                    y=columns[1],
                    hue=label,
                    palette=colors,
                    ax=ax)

    # w_factor = 0.2 / gmm.weights_.max() using for alpha
    for pos, covar, w, color in zip(gmm.means_, gmm.covariances_, gmm.weights_, colors):
        draw_ellipse(pos[:2], covar[:2, :2], ax=ax, alpha=0.1, color=color)

    put.set_title(ax=ax, title=title, **kwargs)
    put.set_ax_ticks_format(ax=ax, xvar_format=var_format, yvar_format=var_format, **kwargs)
    put.set_ax_xy_labels(ax=ax, xlabel=columns[0], ylabel=columns[1], **kwargs)
    for label_, color in zip(ax.legend().get_texts(), colors):
        label_.set_color(color)
        label_.set_size(12)
    ax.get_legend().set_title(label, prop={'size': 12})


class UnitTests(Enum):
    TEST1 = 1
    TEST2 = 2


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.TEST1:
        size = 1000
        p1 = 0.8
        mu1, sigma1 = 0.0, 0.2
        mu2, sigma2 = -1, 0.3
        w = bernoulli.rvs(p1, size=size)

        x1 = np.random.normal(mu1, sigma1, size=size)
        x2 = np.random.normal(mu2, sigma2, size=size)
        x = np.zeros(size)
        for n_ in range(size):
            x[n_] = x1[n_] if w[n_] == 1 else x2[n_]
        x = x.reshape(-1, 1)

        print(np.square(np.std(x1, axis=0)))
        print(np.square(np.std(x2, axis=0)))

        params = fit_gaussian_mixture(x=x)
        print(params)
        plot_mixure1(x=x)

    elif unit_test == UnitTests.TEST2:
        size = 1000
        p1 = 0.8
        mu1, sigma1 = np.array([0, 0]), np.array([[0.2, 0.0], [0.0, 0.2]])
        mu2, sigma2 = np.array([-1.0, 1.00]), np.array([[0.1, 0.0], [0.0, 0.1]])
        w = bernoulli.rvs(p1, size=size)

        x1 = np.random.multivariate_normal(mu1, sigma1, size=size)
        x2 = np.random.multivariate_normal(mu2, sigma2, size=size)
        x = np.zeros((size, 2))
        for n_ in range(size):
            x[n_] = x1[n_] if w[n_] == 1 else x2[n_]

        print(np.square(np.std(x1, axis=0)))
        print(np.square(np.std(x2, axis=0)))

        params = fit_gaussian_mixture(x=x)
        print(params)
        plot_mixure2(x)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.TEST1

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



# packages
import numpy as np
import pandas as pd
import cvxpy as cvx
import seaborn as sns
import matplotlib.pyplot as plt
from numba import jit
from enum import Enum
from typing import Tuple, Optional


class PortfolioObjective(Enum):
    MIN_VAR = 1  # min w^t @ covar @ w
    MAX_LOG_UTIL = 2  # max means^t*w- 0.5*gamma*w^t*covar*w
    EQUAL_RISK_CONTRIBUTION = 3  # implementation in risk_parity
    RISK_PARITY_ALT = 4  # alternative implementation of risk_parity
    MAX_DIVERSIFICATION = 5


def maximize_portfolio_objective_qp(portfolio_objective: PortfolioObjective,
                                    covar: np.ndarray,
                                    means: np.ndarray = None,
                                    weight_mins: np.ndarray = None,
                                    weight_maxs: np.ndarray = None,
                                    is_gross_notional_one: bool = True,
                                    is_long_only: bool = True,
                                    exposure_budget_eq: Optional[Tuple[np.ndarray, float]] = None,
                                    gamma: float = 1.0
                                    ) -> np.ndarray:
    """
    cvx solution for max objective
    subject to linear constraints
         1. weight_min <= w <= weight_max
         2. sum(w) = 1: is_gross_notional_one true or false
         3. exposure_budget_eq[0]^t*w = exposure_budget_eq[1]
    """

    n = covar.shape[0]
    w = cvx.Variable(n)
    portfolio_var = cvx.quad_form(w, covar)

    if portfolio_objective == PortfolioObjective.MIN_VAR:
        objective_fun = -portfolio_var

    elif portfolio_objective == PortfolioObjective.MAX_LOG_UTIL:
        if means is None:
            raise ValueError(f"means must be given")
        objective_fun = means.T @ w - 0.5 * gamma * portfolio_var

    else:
        raise ValueError(f"unknown portfolio_objective")

    objective = cvx.Maximize(objective_fun)

    # add constraints
    constraints = []
    if is_gross_notional_one:
        constraints = constraints + [cvx.sum(w) == 1]
    if is_long_only:
        constraints = constraints + [w >= 0]
    if weight_mins is not None:
        constraints = constraints + [w >= weight_mins]
    if weight_maxs is not None:
        constraints = constraints + [w <= weight_maxs]
    if exposure_budget_eq is not None:
        constraints = constraints + [exposure_budget_eq[0] @ w == exposure_budget_eq[1]]

    problem = cvx.Problem(objective, constraints)
    problem.solve()

    optimal_weights = w.value
    if optimal_weights is None:
        raise ValueError(f"not solved")

    return optimal_weights


def max_qp_portfolio_vol_target(portfolio_objective: PortfolioObjective,
                                covar: np.ndarray,
                                means: np.ndarray = None,
                                weight_min: np.ndarray = None,
                                weight_max: np.ndarray = None,
                                is_gross_notional_one: bool = True,
                                is_long_only: bool = True,
                                exposure_budget_eq: Tuple[np.ndarray, float] = None,
                                vol_target: float = 0.12
                                ) -> np.ndarray:

    max_iter = 20
    sol_tol = 10e-6

    def f(lambda_n: float) -> float:
        w_n = maximize_portfolio_objective_qp(portfolio_objective=portfolio_objective,
                                              covar=covar,
                                              means=means,
                                              weight_mins=weight_min,
                                              weight_maxs=weight_max,
                                              is_gross_notional_one=is_gross_notional_one,
                                              is_long_only=is_long_only,
                                              exposure_budget_eq=exposure_budget_eq,
                                              gamma=lambda_n)

        print('lambda_n='+str(lambda_n))
        print_portfolio_outputs(optimal_weights=w_n,
                                covar=covar,
                                means=means)
        target = w_n.T @ covar @ w_n - vol_target**2
        return target

    # find initials
    cov_inv = np.linalg.inv(covar)
    e = np.ones(covar.shape[0])

    if means is not None:
        a = np.sqrt(e.T@cov_inv@e/(2*vol_target**2))
        b = np.sqrt(means.T@cov_inv@means/(2*vol_target**2))
    else:
        a = np.sqrt(e.T@cov_inv@e/(2*vol_target ** 2))
        b = 100
    f_a = f(a)
    f_b = f(b)

    print((f"initial: {[f_a, f_b]}"))
    if np.sign(f_a) == np.sign(f_b):
        raise ValueError(f"the same signs: {[f_a, f_b]}")

    lambda_n = 0.5 * (a + b)
    for it in range(max_iter):
        lambda_n = 0.5 * (a + b) #new midpoint
        f_n = f(lambda_n)

        if (np.abs(f_n) <= sol_tol) or (np.abs((b-a)/2.0) < sol_tol):
            break
        if np.sign(f_n) == np.sign(f_a):
            a = lambda_n
            f_a = f_n
        else:
            b = lambda_n
        print('it='+str(it))

    w_n = maximize_portfolio_objective_qp(portfolio_objective=portfolio_objective,
                                          covar=covar,
                                          means=means,
                                          weight_mins=weight_min,
                                          weight_maxs=weight_max,
                                          is_gross_notional_one=is_gross_notional_one,
                                          is_long_only=is_long_only,
                                          exposure_budget_eq=exposure_budget_eq,
                                          gamma=lambda_n)
    print_portfolio_outputs(optimal_weights=w_n,
                            covar=covar,
                            means=means)
    return w_n


def max_portfolio_sharpe_qp(covar: np.ndarray,
                            means: np.ndarray,
                            weight_mins: np.ndarray = None,
                            weight_maxs: np.ndarray = None,
                            is_gross_notional_one: bool = False,
                            is_long_only: bool = True,
                            exposure_budget_eq: Tuple[np.ndarray, float] = None,
                            exposure_budget_le: Tuple[np.ndarray, float] = None,
                            is_print_log: bool = True
                            ) -> np.ndarray:
    """
    max means^t*w / sqrt(w^t @ covar @ w)
    subject to
     1. weight_min <= w <= weight_max
    """
    n = covar.shape[0]
    z = cvx.Variable(n+1)
    y = z[:n]
    k = z[n]

    objective = cvx.Minimize(cvx.quad_form(y, covar))

    # add constraints
    constraints = [means.T @ y == 1.0]  # scaling

    if is_gross_notional_one:
        constraints = constraints + [cvx.sum(y) == k]
#    else:
#        constraints = constraints + [cvx.sum(y) >= k]  #scaling

    if is_long_only:
        constraints = constraints + [y >= 0]

    if exposure_budget_eq is not None:
        constraints = constraints + [exposure_budget_eq[0] @ y == exposure_budget_eq[1]*k]

    if exposure_budget_le is not None:
        constraints = constraints + [exposure_budget_le[0] @ y <= exposure_budget_le[1]*k]

    if weight_mins is not None:
        constraints = constraints + [y >= k * weight_mins]
    if weight_maxs is not None:
        constraints = constraints + [y <= k * weight_maxs]

    problem = cvx.Problem(objective, constraints)
    problem.solve()

    optimal_weights = z.value

    if optimal_weights is not None:
        optimal_weights = optimal_weights[:n] / optimal_weights[n]  # apply rescaling
    else:

        print(f"max_port_sharpe_qp optimal_weights not found = {optimal_weights}")
        print(f"covar={covar}")
        print(f"means={means}")
        if weight_mins is not None:
            print(f"setting optimal_weights to weight_min = {weight_mins}")
            optimal_weights = weight_mins
        else:
            print(f"setting optimal_weights to 0")
            optimal_weights = np.zeros_like(n)

    if is_print_log:
        weights_str = ' '.join([f"{w:0.3f}" for w in optimal_weights])
        print(weights_str)

    return optimal_weights


@jit(nopython=True)
def solve_analytic_log_opt(covar: np.ndarray,
                           means: np.ndarray,
                           exposure_budget_eq: Tuple[np.ndarray, float] = None,
                           gamma: float = 1.0
                           )-> np.ndarray:

    """
    analytic solution for max{means^t*w - 0.5*gamma*w^t*covar*w}
    subject to exposure_budget_eq[0]^t*w = exposure_budget_eq[1]
    """
    sigma_i = np.linalg.inv(covar)

    if exposure_budget_eq is not None:

        # get constraints
        a = exposure_budget_eq[0]
        # if len(a) != covar.shape[0]:
            # raise ValueError(f"dimensions of exposure constraint {a} not matichng covar dimensions")
        a0 = exposure_budget_eq[1]
        # if not isinstance(a0, float):
            # raise ValueError(f"a0 = {a0} must be single float")

        a_sigma_a = a.T @ sigma_i @ a
        a_sigma_mu = a.T @ sigma_i @ means
        l_lambda = (-gamma*a0+a_sigma_mu) / a_sigma_a
        optimal_weights = (1.0/gamma) * sigma_i @ (means - l_lambda * a)

    else:
        optimal_weights = (1.0/gamma) * sigma_i @ means

    return optimal_weights


def print_portfolio_outputs(optimal_weights: np.ndarray,
                            covar: np.ndarray,
                            means: np.ndarray) -> None:

    mean = means.T @ optimal_weights
    vol = np.sqrt(optimal_weights.T @ covar @ optimal_weights)
    sharpe = mean / vol
    inst_sharpes = means / np.sqrt(np.diag(covar))
    sharpe_weighted = inst_sharpes.T @ (optimal_weights / np.sum(optimal_weights))

    line_str = (f"expected={mean: 0.2%}, "
                f"vol={vol: 0.2%}, "
                f"Sharpe={sharpe: 0.2f}, "
                f"weighted Sharpe={sharpe_weighted: 0.2f}, "
                f"inst Sharpes={np.array2string(inst_sharpes, precision=2)}, "
                f"weights={np.array2string(optimal_weights, precision=2)}")

    print(line_str)


class UnitTests(Enum):
    MIN_VAR = 1
    MAX_UTILITY = 2
    EFFICIENT_FRONTIER = 3
    MAX_UTILITY_VOL_TARGET = 4
    SHARPE = 5
    REGIME_SHARPE = 6


def run_unit_test(unit_test: UnitTests):

    means = np.array([-0.01, 0.05])  # sharpe = [-.1, 0.5]
    covar = np.array([[0.2**2, -0.0075],
                      [-0.0075, 0.1**2]])

    if unit_test == UnitTests.MIN_VAR:

        weight_min = np.array([0.0, 0.0])
        weight_max = np.array([10.0, 10.0])

        optimal_weights = maximize_portfolio_objective_qp(portfolio_objective=PortfolioObjective.MIN_VAR,
                                                          covar=covar,
                                                          means=means,
                                                          weight_mins=None,
                                                          weight_maxs=None,
                                                          is_gross_notional_one=True,
                                                          is_long_only=True,
                                                          exposure_budget_eq=None)

        print_portfolio_outputs(optimal_weights=optimal_weights,
                                covar=covar,
                                means=means)

    elif unit_test == UnitTests.MAX_UTILITY:

        gamma = 50*np.trace(covar)
        optimal_weights = maximize_portfolio_objective_qp(portfolio_objective=PortfolioObjective.MAX_LOG_UTIL,
                                                          covar=covar,
                                                          means=means,
                                                          weight_mins=None,
                                                          weight_maxs=None,
                                                          is_gross_notional_one=False,
                                                          is_long_only=True,
                                                          exposure_budget_eq=None,
                                                          gamma=gamma)

        print_portfolio_outputs(optimal_weights=optimal_weights,
                                covar=covar,
                                means=means)

    elif unit_test == UnitTests.EFFICIENT_FRONTIER:

        portfolio_mus = []
        portfolio_vols = []
        portfolio_sharpes = []
        w_lambdas = []
        lang_lambdas = np.arange(0.5, 100.0, 1.0)
        exposure_budget_eq = (np.ones_like(means), 1.0)

        for lang_lambda in lang_lambdas:
            is_analytic = False
            if is_analytic:
                w_lambda = solve_analytic_log_opt(covar=covar,
                                                  means=means,
                                                  exposure_budget_eq=exposure_budget_eq,
                                                  gamma=lang_lambda)
            else:
                w_lambda = maximize_portfolio_objective_qp(portfolio_objective=PortfolioObjective.MAX_LOG_UTIL,
                                                           covar=covar,
                                                           means=means,
                                                           is_gross_notional_one=True,
                                                           gamma=lang_lambda)

            portfolio_vol = np.sqrt(w_lambda.T@covar@w_lambda)
            portfolio_sharpe = means.T @ w_lambda / portfolio_vol
            portfolio_mus.append(means.T @ w_lambda)
            w_lambdas.append(w_lambda)
            portfolio_vols.append(portfolio_vol)
            portfolio_sharpes.append(portfolio_sharpe)

        portfolio_return = pd.Series(portfolio_mus, index=lang_lambdas).rename('mean')
        portfolio_vol = pd.Series(portfolio_vols, index=lang_lambdas).rename('vol')
        portfolio_sharpe = pd.Series(portfolio_sharpes, index=lang_lambdas).rename('Sharpe')
        w_lambdas = pd.DataFrame(w_lambdas, index=lang_lambdas)
        protfolio_data = pd.concat([portfolio_return, portfolio_vol, portfolio_sharpe, w_lambdas], axis=1)
        print(protfolio_data)
        fig, axs = plt.subplots(2, 1, figsize=(15, 12))
        sns.lineplot(x='vol', y='mean', data=protfolio_data, ax=axs[0])
        sns.lineplot(data=protfolio_data[['mean', 'vol']], ax=axs[1])

    elif unit_test == UnitTests.MAX_UTILITY_VOL_TARGET:

        optimal_weights = max_qp_portfolio_vol_target(portfolio_objective=PortfolioObjective.MAX_LOG_UTIL,
                                                      covar=covar,
                                                      means=means,
                                                      weight_min=None,
                                                      weight_max=None,
                                                      is_gross_notional_one=True,
                                                      is_long_only=True,
                                                      exposure_budget_eq=None,
                                                      vol_target=0.08)

        print_portfolio_outputs(optimal_weights=optimal_weights,
                                covar=covar,
                                means=means)

    elif unit_test == UnitTests.SHARPE:

        portfolio_mus = []
        portfolio_vols = []
        portfolio_sharpes = []
        exposure_budget_eq = (np.ones_like(means), 1.0)

        lang_lambdas = np.arange(1.0, 20.0, 1.0)
        for lang_lambda in lang_lambdas:
            is_analytic = True
            if is_analytic:
                w_lambda = solve_analytic_log_opt(covar=covar,
                                                  means=means,
                                                  exposure_budget_eq=exposure_budget_eq,
                                                  gamma=lang_lambda)
            else:
                w_lambda = maximize_portfolio_objective_qp(portfolio_objective=PortfolioObjective.MAX_LOG_UTIL,
                                                           covar=covar,
                                                           means=means,
                                                           exposure_budget_eq=exposure_budget_eq,
                                                           gamma=lang_lambda)

            print(f"portfolio with lambda = {lang_lambda}")
            print_portfolio_outputs(optimal_weights=w_lambda,
                                    covar=covar,
                                    means=means)

            portfolio_vol = np.sqrt(w_lambda.T@covar@w_lambda)
            portfolio_sharpe = means.T @ w_lambda / portfolio_vol
            portfolio_mus.append(means.T @ w_lambda)
            portfolio_vols.append(portfolio_vol)
            portfolio_sharpes.append(portfolio_sharpe)

        portfolio_return = pd.Series(portfolio_mus, index=lang_lambdas)
        portfolio_vol = pd.Series(portfolio_vols, index=lang_lambdas)
        portfolio_sharpe = pd.Series(portfolio_sharpes, index=lang_lambdas)
        protfolio_data = pd.concat([portfolio_return, portfolio_vol, portfolio_sharpe], axis=1)
        print(protfolio_data)

        opt_sharpe_w = max_portfolio_sharpe_qp(covar=covar,
                                               means=means,
                                               exposure_budget_eq=exposure_budget_eq)

        print(f"exact solution")
        print_portfolio_outputs(optimal_weights=opt_sharpe_w,
                                covar=covar,
                                means=means)

    elif unit_test == UnitTests.REGIME_SHARPE:

        # case of two assets:
        # inputs:
        g = 3

        # individual
        sharpes = np.array((0.4, 0.3))
        betas_port = np.array((1.0, 0.8, 1.0))
        betas_cta = np.array((-1.0, 0.25, 0.25))
        idio_vols = np.array((0.01, 0.1))

        # factor
        p_regimes = np.array((0.16, 0.68, 0.16))
        factor_vol = 0.15

        betas_matrix = np.stack((betas_port, betas_cta))
        print(betas_matrix)

        n = betas_matrix.shape[0]
        covar = np.zeros((n, n))
        for g_ in range(g):
            b = betas_matrix[:, g_]
            covar_regime = np.outer(b, b)
            print(f"covar_regime=\n{covar_regime}")
            covar += covar_regime*p_regimes[g_]

        covar = (factor_vol**2) * covar + np.diag(idio_vols**2)
        print(f"t_covar_regime=\n{covar}")

        implied_vols = np.sqrt(np.diag(covar))
        print(f"implied_vols=\n{implied_vols}")

        means = sharpes * implied_vols
        print(f"implied_means=\n{means}")

        # invest 100% in first asset
        exposure_budget_eq = (np.array([1.0, 0.0]), 1.0)
        optimal_weights = max_portfolio_sharpe_qp(covar=covar,
                                                  means=means,
                                                  exposure_budget_eq=exposure_budget_eq)

        print_portfolio_outputs(optimal_weights=optimal_weights,
                                covar=covar,
                                means=means)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.SHARPE

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)
